import dataclasses
import logging
import warnings
from collections.abc import Mapping
from typing import Any, Optional

import grpc
import grpc_reflection
import proto.message
from google._upb._message import DescriptorPool
from google.protobuf import (
    descriptor_pb2,
    json_format,
    message_factory,
    symbol_database,
)
from google.protobuf.json_format import ParseError
from grpc_reflection.v1alpha import reflection_pb2, reflection_pb2_grpc
from grpc_status import rpc_status

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys
from tavern._plugins.grpc.protos import _generate_proto_import, _import_grpc_module

logger: logging.Logger = logging.getLogger(__name__)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    warnings.warn("deprecated", DeprecationWarning)  # noqa: B028

_ProtoMessageType = type[proto.message.Message]


@dataclasses.dataclass
class _ChannelVals:
    channel: grpc.UnaryUnaryMultiCallable
    input_type: _ProtoMessageType
    output_type: _ProtoMessageType


class GRPCClient:
    def __init__(self, **kwargs) -> None:
        logger.debug("Initialising GRPC client with %s", kwargs)
        expected_blocks = {
            "connect": {"host", "port", "options", "timeout", "secure"},
            "proto": {"source", "module"},
            "metadata": {},
            "attempt_reflection": {},
        }
        # check main block first
        check_expected_keys(expected_blocks.keys(), kwargs)

        _connect_args = kwargs.pop("connect", {})
        check_expected_keys(expected_blocks["connect"], _connect_args)

        metadata = kwargs.pop("metadata", {})
        self._metadata = list(metadata.items())

        _proto_args = kwargs.pop("proto", {})
        check_expected_keys(expected_blocks["proto"], _proto_args)

        self._attempt_reflection = bool(kwargs.pop("attempt_reflection", False))

        if default_host := _connect_args.get("host"):
            self.default_host = default_host
            if port := _connect_args.get("port"):
                self.default_host += f":{port}"

        self.timeout = int(_connect_args.get("timeout", 5))
        self.secure = bool(_connect_args.get("secure", False))

        self._options: list[tuple[str, Any]] = []
        for key, value in _connect_args.pop("options", {}).items():
            if not key.startswith("grpc."):
                raise exceptions.GRPCServiceException(
                    f"invalid grpc option '{key}' - must be in the form 'grpc.option_name'"
                )
            self._options.append((key, value))

        self.channels: dict[str, grpc.Channel] = {}
        # Using the default symbol database is a bit undesirable because it means that things being imported from
        # previous tests will affect later ones which can mask bugs. But there isn't a nice way to have a
        # self-contained symbol database, because then you need to transitively import all dependencies of protos and
        # add them to the database.
        self.sym_db = symbol_database.Default()

        if proto_source := _proto_args.get("source"):
            _generate_proto_import(proto_source)

        if proto_module := _proto_args.get("module"):
            try:
                _import_grpc_module(proto_module)
            except (ImportError, ModuleNotFoundError) as e:
                logger.exception(f"could not import {proto_module}")
                raise exceptions.GRPCServiceException(
                    "error importing gRPC modules"
                ) from e

    def _register_file_descriptor(
        self,
        service_proto: grpc_reflection.v1alpha.reflection_pb2.FileDescriptorResponse,
    ) -> None:
        for file_descriptor_proto in service_proto.file_descriptor_proto:
            descriptor = descriptor_pb2.FileDescriptorProto()
            descriptor.ParseFromString(file_descriptor_proto)
            self.sym_db.pool.Add(descriptor)

    def _get_reflection_info(
        self, channel, service_name: Optional[str] = None, file_by_filename=None
    ) -> None:
        logger.debug(
            "Getting GRPC protobuf for service %s from reflection", service_name
        )
        ref_request = reflection_pb2.ServerReflectionRequest(
            file_containing_symbol=service_name, file_by_filename=file_by_filename
        )
        reflection_stub = reflection_pb2_grpc.ServerReflectionStub(channel)
        ref_response = reflection_stub.ServerReflectionInfo(
            iter([ref_request]), metadata=self._metadata
        )
        for response in ref_response:
            self._register_file_descriptor(response.file_descriptor_response)

    def _get_grpc_service(
        self, channel: grpc.Channel, service: str, method: str
    ) -> Optional[_ChannelVals]:
        full_service_name = f"{service}/{method}"
        try:
            input_type, output_type = self.get_method_types(full_service_name)
        except KeyError as e:
            logger.debug(f"could not find types: {e}")
            return None

        logger.info(f"reflected info for {service}: {full_service_name}")

        grpc_method = channel.unary_unary(
            "/" + full_service_name,
            request_serializer=input_type.SerializeToString,
            response_deserializer=output_type.FromString,
        )

        return _ChannelVals(grpc_method, input_type, output_type)

    def get_method_types(
        self, full_method_name: str
    ) -> tuple[_ProtoMessageType, _ProtoMessageType]:
        """Uses the builtin symbol pool to try and find the input and output types for the given method

        Args:
            full_method_name: full RPC name in the form 'pkg.ServiceName/Method'

        Returns:
            input and output types (class objects) for the RPC

        Raises:
            KeyError: If the types are not registered. Should ideally never happen?
        """
        logger.debug(f"looking up types for {full_method_name}")

        service, method = full_method_name.split("/")

        pool: DescriptorPool = self.sym_db.pool
        grpc_service = pool.FindServiceByName(service)
        method = grpc_service.FindMethodByName(method)
        input_type = message_factory.GetMessageClass(method.input_type)  # type: ignore
        output_type = message_factory.GetMessageClass(method.output_type)  # type: ignore

        return input_type, output_type

    def _make_call_request(
        self, host: str, full_service: str
    ) -> Optional[_ChannelVals]:
        full_service = full_service.replace("/", ".")
        service_method = full_service.rsplit(".", 1)
        if len(service_method) != 2:
            raise exceptions.GRPCRequestException(
                f"Invalid service/method name {full_service}"
            )

        service = service_method[0]
        method = service_method[1]
        logger.debug(
            "Make call for host %s service %s method %s", host, service, method
        )

        if host not in self.channels:
            if self.secure:
                credentials = grpc.ssl_channel_credentials()
                self.channels[host] = grpc.secure_channel(
                    host,
                    credentials,
                    options=self._options,
                )
            else:
                self.channels[host] = grpc.insecure_channel(
                    host,
                    options=self._options,
                )

        channel = self.channels[host]

        if channel_vals := self._get_grpc_service(channel, service, method):
            return channel_vals

        if not self._attempt_reflection:
            logger.error(
                "could not find service and gRPC reflection disabled, cannot continue"
            )
            raise exceptions.GRPCServiceException(
                f"Service {service} was not registered for host {host}"
            )

        logger.info("service not registered, doing reflection from server")
        try:
            self._get_reflection_info(channel, service_name=service)
        except grpc.RpcError as rpc_error:
            code = details = None
            try:
                code = rpc_error.code()
                details = rpc_error.details()
            except AttributeError:
                status = rpc_status.from_call(rpc_error)
                if status is None:
                    logger.warning("Unknown error occurred in RPC call", exc_info=True)
                else:
                    code = status.code
                    details = status.details

            if code and details:
                logger.warning(
                    "Unable get %s service reflection information code %s detail %s",
                    service,
                    code,
                    details,
                    exc_info=True,
                )

            raise exceptions.GRPCRequestException from rpc_error

        return self._get_grpc_service(channel, service, method)

    def __enter__(self) -> "GRPCClient":
        logger.debug("Connecting to GRPC")
        return self

    def call(
        self,
        service: str,
        host: Optional[str] = None,
        body: Optional[Mapping] = None,
        timeout: Optional[int] = None,
    ) -> grpc.Future:
        """Makes the request and returns a future with the response."""
        if host is None:
            if getattr(self, "default_host", None) is None:
                raise exceptions.GRPCRequestException(
                    "no host specified in request and no default host in settings"
                )

            host = self.default_host

        if timeout is None:
            timeout = self.timeout

        channel_vals = self._make_call_request(host, service)
        if channel_vals is None:
            raise exceptions.GRPCServiceException(
                f"Service {service} was not found on host {host}"
            )

        request = channel_vals.input_type()
        if body is not None:
            try:
                request = json_format.ParseDict(body, request)
            except ParseError as e:
                raise exceptions.GRPCRequestException(
                    "error creating request from json body"
                ) from e

        logger.debug("Sending request %s", request)

        return channel_vals.channel.future(
            request, metadata=self._metadata, timeout=timeout
        )

    def __exit__(self, *args) -> None:
        logger.debug("Disconnecting from GRPC")
        for v in self.channels.values():
            v.close()
        self.channels = {}



import logging
import os
import tempfile

import pytest

logger = logging.getLogger(__name__)

name = None


@pytest.fixture(autouse=True)
def setup_logging():
    logging.basicConfig(level=logging.INFO)


def pytest_tavern_beta_after_every_response(expected, response):
    global name
    logging.debug(expected)
    logging.debug(response)
    with open(name, "a") as tfile:
        tfile.write("abc\n")


@pytest.fixture(autouse=True)
def after_check_result():
    """Create a temporary file for the duration of the test, and make sure the above hook was called"""
    global name
    with tempfile.NamedTemporaryFile(delete=False) as tfile:
        try:
            tfile.close()

            name = tfile.name
            yield

            with open(tfile.name, "r") as opened:
                contents = opened.readlines()
                assert len(contents) == 2
                assert all(i.strip() == "abc" for i in contents)

        finally:
            os.remove(tfile.name)



import logging

import requests

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.plugins import PluginHelperBase
from tavern._core.pytest.config import TestConfig

from .request import RestRequest
from .response import RestResponse

logger: logging.Logger = logging.getLogger(__name__)


class TavernRestPlugin(PluginHelperBase):
    session_type = requests.Session

    request_type = RestRequest
    request_block_name = "request"

    @staticmethod
    def get_expected_from_request(
        response_block: dict, test_block_config: TestConfig, session
    ):
        if response_block is None:
            raise exceptions.MissingSettingsError(
                "no response block specified for HTTP test stage"
            )

        f_expected = format_keys(response_block, test_block_config.variables)
        return f_expected

    verifier_type = RestResponse
    response_block_name = "response"



import functools
import json
import logging

from box.box import Box

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys, format_keys
from tavern._core.extfunctions import update_from_ext
from tavern._core.pytest.config import TestConfig
from tavern._core.report import attach_yaml
from tavern._plugins.mqtt.client import MQTTClient
from tavern.request import BaseRequest

logger: logging.Logger = logging.getLogger(__name__)


def get_publish_args(rspec: dict, test_block_config: TestConfig) -> dict:
    """Format mqtt request args and update using ext functions"""

    fspec = format_keys(rspec, test_block_config.variables)

    if "json" in fspec:
        if "payload" in fspec:
            raise exceptions.BadSchemaError(
                "Can only specify one of 'payload' or 'json' in MQTT request"
            )

        update_from_ext(fspec, ["json"])

        fspec["payload"] = json.dumps(fspec.pop("json"))

    return fspec


class MQTTRequest(BaseRequest):
    """Wrapper for a single mqtt request on a client

    Similar to RestRequest, publishes a single message.
    """

    def __init__(
        self, client: MQTTClient, rspec: dict, test_block_config: TestConfig
    ) -> None:
        expected = {"topic", "payload", "json", "qos", "retain"}

        check_expected_keys(expected, rspec)

        publish_args = get_publish_args(rspec, test_block_config)

        self._publish_args = publish_args
        self._prepared = functools.partial(client.publish, **publish_args)

        # Need to do this here because get_publish_args will modify the original
        # input, which we might want to use to format. No error handling because
        # all the error handling is done in the previous call
        self._original_publish_args = format_keys(rspec, test_block_config.variables)

        # TODO
        # From paho:
        # > raise TypeError('payload must be a string, bytearray, int, float or None.')
        # Need to be able to take all of these somehow, and also match these
        # against any payload received on the topic

    def run(self):
        attach_yaml(
            self._original_publish_args,
            name="rest_request",
        )

        try:
            return self._prepared()
        except ValueError as e:
            logger.exception("Error publishing")
            raise exceptions.MQTTRequestException from e

    @property
    def request_vars(self) -> Box:
        return Box(self._original_publish_args)



import logging.config
import os

import pytest
import stevedore
import yaml

import tavern
import tavern._plugins.mqtt.tavernhook as mqtt_plugin
from tavern._plugins.rest.tavernhook import TavernRestPlugin as rest_plugin


@pytest.fixture(scope="function", autouse=True)
def run_all():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(current_dir, "logging.yaml")) as spec_file:
        settings = yaml.load(spec_file, Loader=yaml.SafeLoader)
        logging.config.dictConfig(settings)


@pytest.fixture(scope="session", autouse=True)
def set_plugins():
    def extension(name, point):
        return stevedore.extension.Extension(name, point, point, point)

    tavern._core.plugins.load_plugins.plugins = [
        extension(
            "requests",
            rest_plugin,
        ),
        extension(
            "paho-mqtt",
            mqtt_plugin,
        ),
    ]



import contextlib
import json
import logging
import warnings
from collections.abc import Mapping
from contextlib import ExitStack
from itertools import filterfalse, tee
from typing import Callable, ClassVar, Optional
from urllib.parse import quote_plus

import requests
from box.box import Box
from requests.cookies import cookiejar_from_dict
from requests.utils import dict_from_cookiejar

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys, deep_dict_merge, format_keys
from tavern._core.extfunctions import update_from_ext
from tavern._core.general import valid_http_methods
from tavern._core.pytest.config import TestConfig
from tavern._core.report import attach_yaml
from tavern._plugins.rest.files import get_file_arguments, guess_filespec
from tavern.request import BaseRequest

logger: logging.Logger = logging.getLogger(__name__)


def get_request_args(rspec: dict, test_block_config: TestConfig) -> dict:
    """Format the test spec given values inthe global config

    Todo:
        Add similar functionality to validate/save $ext functions so input
        can be generated from a function

    Args:
        rspec: Test spec
        test_block_config: Test block config

    Returns:
        Formatted test spec

    Raises:
        BadSchemaError: Tried to pass a body in a GET request
    """

    if "method" not in rspec:
        logger.debug("Using default GET method")
        rspec["method"] = "GET"

    if "headers" not in rspec:
        rspec["headers"] = {}

    content_keys = ["data", "json", "files", "file_body"]

    in_request = [c for c in content_keys if c in rspec]
    if len(in_request) > 1:
        # Explicitly raise an error here
        # From requests docs:
        # Note, the json parameter is ignored if either data or files is passed.
        # However, we allow the data + files case, as requests handles it correctly
        if set(in_request) != {"data", "files"}:
            raise exceptions.BadSchemaError(
                "Can only specify one type of request data in HTTP request (tried to "
                "send {})".format(" and ".join(in_request))
            )

    normalised_headers = {k.lower(): v for k, v in rspec["headers"].items()}

    def get_header(name):
        return normalised_headers.get(name, None)

    content_header = get_header("content-type")
    encoding_header = get_header("content-encoding")

    if "files" in rspec:
        if content_header:
            logger.warning(
                "Tried to specify a content-type header while sending multipart files - this will be ignored"
            )
            rspec["headers"] = {
                i: j
                for i, j in normalised_headers.items()
                if i.lower() != "content-type"
            }

    fspec = format_keys(rspec, test_block_config.variables)

    if fspec["method"] not in valid_http_methods:
        raise exceptions.BadSchemaError(
            "Unknown HTTP method {}".format(fspec["method"])
        )

    # If the user is using the file_body key, try to guess what type of file/encoding it is.
    filename = fspec.get("file_body")
    if filename:
        with ExitStack() as stack:
            file_spec, group_name = guess_filespec(filename, stack, test_block_config)

            # Group name doesn't matter here as it's a single file
            if group_name:
                logger.warning(
                    f"'group_name' for the 'file_body' key was specified as '{group_name}' but this will be ignored "
                )

            fspec["file_body"] = filename
            if len(file_spec) == 2:
                logger.debug(
                    "No content type or encoding inferred from file_body for %s",
                    filename,
                )

            if len(file_spec) >= 3:
                inferred_content_type = file_spec[2]
                if content_header:
                    logger.info(
                        "inferred content type '%s' from %s, but using user specified content type '%s'",
                        inferred_content_type,
                        filename,
                        content_header,
                    )
                else:
                    fspec["headers"]["content-type"] = inferred_content_type

            if len(file_spec) == 4:
                inferred_content_encoding = file_spec[3]
                if encoding_header:
                    logger.info(
                        "inferred content encoding '%s' from %s, but using user specified encoding '%s",
                        inferred_content_encoding,
                        filename,
                        encoding_header,
                    )
                else:
                    fspec["headers"].update(**inferred_content_encoding)

    #########################################

    request_args = {}

    def add_request_args(keys, optional):
        for key in keys:
            try:
                request_args[key] = fspec[key]
            except KeyError:
                if optional or (key in request_args):
                    continue

                # This should never happen
                raise

    # Ones that are required and are enforced to be present by the schema
    required_in_file = ["method", "url"]

    add_request_args(["file_body"], True)
    add_request_args(required_in_file, False)
    add_request_args(RestRequest.optional_in_file, True)

    if "auth" in fspec:
        request_args["auth"] = tuple(fspec["auth"])

    if "cert" in fspec:
        if isinstance(fspec["cert"], list):
            request_args["cert"] = tuple(fspec["cert"])

    if "timeout" in fspec:
        # Needs to be a tuple, it being a list doesn't work
        if isinstance(fspec["timeout"], list):
            request_args["timeout"] = tuple(fspec["timeout"])

    # If there's any nested json in parameters, urlencode it
    # if you pass nested json to 'params' then requests silently fails and just
    # passes the 'top level' key, ignoring all the nested json. I don't think
    # there's a standard way to do this, but urlencoding it seems sensible
    # eg https://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter
    # > ...represented in an OAuth 2.0 request as UTF-8 encoded JSON (which ends
    # > up being form-urlencoded when passed as an OAuth parameter)
    for key, value in request_args.get("params", {}).items():
        if not isinstance(value, str):
            if key == "$ext":
                logger.debug("Skipping converting of ext function (%s)", value)
                continue

        if isinstance(value, dict):
            request_args["params"][key] = quote_plus(json.dumps(value))

    optional = {"verify", "stream"}

    for key in optional:
        if key in fspec:
            request_args[key] = fspec[key]

    # TODO
    # requests takes all of these - we need to parse the input to get them
    # "cookies",

    # These verbs _can_ send a body but the body _should_ be ignored according
    # to the specs - some info here:
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
    if request_args["method"] in ["GET", "HEAD", "OPTIONS"]:
        if any(i in request_args for i in ["json", "data"]):
            warnings.warn(  # noqa
                "You are trying to send a body with a HTTP verb that has no semantic use for it",
                RuntimeWarning,
            )

    return request_args


@contextlib.contextmanager
def _set_cookies_for_request(session: requests.Session, request_args: Mapping):
    """
    Possibly reset session cookies for a single request then set them back.
    If no cookies were present in the request arguments, do nothing.

    This does not use try/finally because if it fails then we don't care about
    the cookies anyway

    Args:
        session: Current session
        request_args: current request arguments
    """
    if "cookies" in request_args:
        old_cookies = dict_from_cookiejar(session.cookies)
        session.cookies = cookiejar_from_dict({})
        yield
        session.cookies = cookiejar_from_dict(old_cookies)
    else:
        yield


def _check_allow_redirects(rspec: dict, test_block_config: TestConfig):
    """
    Check for allow_redirects flag in settings/stage

    Args:
        rspec: request dictionary
        test_block_config: config available for test

    Returns:
        Whether to allow redirects for this stage or not
    """
    # By default, don't follow redirects
    allow_redirects = False

    # Then check to see if we should follow redirects based on settings
    global_follow_redirects = test_block_config.follow_redirects
    if global_follow_redirects is not None:
        allow_redirects = global_follow_redirects

    # ... and test flags
    test_follow_redirects = rspec.pop("follow_redirects", None)
    if test_follow_redirects is not None:
        if global_follow_redirects is not None:
            logger.info(
                "Overriding global follow_redirects setting of %s with test-level specification of %s",
                global_follow_redirects,
                test_follow_redirects,
            )
        allow_redirects = test_follow_redirects

    logger.debug("Allow redirects in stage: %s", allow_redirects)

    return allow_redirects


def _read_expected_cookies(
    session: requests.Session, rspec: Mapping, test_block_config: TestConfig
) -> Optional[dict]:
    """
    Read cookies to inject into request, ignoring others which are present

    Args:
        session: session object
        rspec: test spec
        test_block_config: config available for test

    Returns:
        cookies to use in request, if any
    """
    # Need to do this down here - it is separate from getting request args as
    # it depends on the state of the session
    existing_cookies = session.cookies.get_dict()
    cookies_to_use = format_keys(
        rspec.get("cookies", None), test_block_config.variables
    )

    if cookies_to_use is None:
        logger.debug("No cookies specified in request, sending all")
        return None
    elif cookies_to_use in ([], {}):
        logger.debug("Not sending any cookies with request")
        return {}

    def partition(pred, iterable):
        """From itertools documentation"""
        t1, t2 = tee(iterable)
        return list(filterfalse(pred, t1)), list(filter(pred, t2))

    # Cookies are either a single list item, specitying which cookie to send, or
    # a mapping, specifying cookies to override
    expected, extra = partition(lambda x: isinstance(x, dict), cookies_to_use)

    missing = set(expected) - set(existing_cookies.keys())

    if missing:
        logger.error("Missing cookies")
        raise exceptions.MissingCookieError(
            "Tried to use cookies '{}' in request but only had '{}' available".format(
                expected, existing_cookies
            )
        )

    # 'extra' should be a list of dictionaries - merge them into one here
    from_extra = {k: v for mapping in extra for (k, v) in mapping.items()}

    if len(extra) != len(from_extra):
        logger.error("Duplicate cookie override values specified")
        raise exceptions.DuplicateCookieError(
            "Tried to override the value of a cookie multiple times in one request"
        )

    overwritten = [i for i in expected if i in from_extra]

    if overwritten:
        logger.error("Duplicate cookies found in request")
        raise exceptions.DuplicateCookieError(
            "Asked to use cookie {} from previous request but also redefined it as {}".format(
                overwritten, from_extra
            )
        )

    from_cookiejar = {c: existing_cookies.get(c) for c in expected}

    return deep_dict_merge(from_cookiejar, from_extra)


class RestRequest(BaseRequest):
    optional_in_file: ClassVar[list[str]] = [
        "json",
        "data",
        "params",
        "headers",
        "files",
        "timeout",
        "cert",
        # Ideally this would just be passed through but requests seems to error
        # if we pass a list instead of a tuple, so we have to manually convert
        # it further down
        # "auth"
    ]

    def __init__(
        self, session: requests.Session, rspec: dict, test_block_config: TestConfig
    ) -> None:
        """Prepare request

        Args:
            session: existing session
            rspec: test spec
            test_block_config: Any configuration for this the block of
                tests

        Raises:
            UnexpectedKeysError: If some unexpected keys were used in the test
                spec. Only valid keyword args to requests can be passed
        """

        if rspec.pop("clear_session_cookies", False):
            session.cookies.clear_session_cookies()

        expected = {
            "method",
            "url",
            "headers",
            "data",
            "params",
            "auth",
            "json",
            "verify",
            "files",
            "file_body",
            "stream",
            "timeout",
            "cookies",
            "cert",
            # "hooks",
            "follow_redirects",
        }

        check_expected_keys(expected, rspec)

        request_args = get_request_args(rspec, test_block_config)
        update_from_ext(
            request_args,
            RestRequest.optional_in_file + ["url"],
        )

        # Used further down, but pop it asap to avoid unwanted side effects
        file_body: Optional[str] = request_args.pop("file_body", None)

        # If there was a 'cookies' key, set it in the request
        expected_cookies = _read_expected_cookies(session, rspec, test_block_config)
        if expected_cookies is not None:
            logger.debug("Sending cookies %s in request", expected_cookies.keys())
            request_args.update(cookies=expected_cookies)

        # Check for redirects
        request_args.update(
            allow_redirects=_check_allow_redirects(rspec, test_block_config)
        )

        logger.debug("Request args: %s", request_args)

        self._request_args = request_args

        # There is no way using requests to make a prepared request that will
        # not follow redirects, so instead we have to do this. This also means
        # that we can't have the 'pre-request' hook any more because we don't
        # create a prepared request.

        def prepared_request():
            # If there are open files, create a context manager around each so
            # they will be closed at the end of the request.
            with ExitStack() as stack:
                stack.enter_context(_set_cookies_for_request(session, request_args))

                # These are mutually exclusive
                if file_body:
                    # Any headers will have been set in the above function
                    file = stack.enter_context(open(file_body, "rb"))
                    request_args.update(data=file)
                else:
                    files = get_file_arguments(request_args, stack, test_block_config)
                    if files:
                        logger.debug("Sending %d files in request", len(files["files"]))
                        self._request_args.update(files)

                headers = self._request_args.get("headers", {})
                for k, v in headers.items():
                    headers[str(k)] = str(v)

                return session.request(**self._request_args)

        self._prepared: Callable[[], requests.Response] = prepared_request

    def run(self) -> requests.Response:
        """Runs the prepared request and times it

        Returns:
            response object
        """

        attach_yaml(
            self._request_args,
            name="rest_request",
        )

        try:
            return self._prepared()
        except requests.exceptions.RequestException as e:
            logger.exception("Error running prepared request")
            raise exceptions.RestRequestException from e

    @property
    def request_vars(self) -> Box:
        return Box(self._request_args)



import dataclasses
import logging
import traceback
from abc import abstractmethod
from collections.abc import Mapping
from textwrap import indent
from typing import Any, Optional

from tavern._core import exceptions
from tavern._core.dict_util import check_keys_match_recursive, recurse_access_key
from tavern._core.extfunctions import get_wrapped_response_function
from tavern._core.pytest.config import TestConfig
from tavern._core.strict_util import StrictOption

logger: logging.Logger = logging.getLogger(__name__)


def indent_err_text(err: str) -> str:
    if err == "null":
        err = "<No body>"
    return indent(err, " " * 4)


@dataclasses.dataclass
class BaseResponse:
    name: str
    expected: Any
    test_block_config: TestConfig
    response: Optional[Any] = None

    validate_functions: list[Any] = dataclasses.field(init=False, default_factory=list)
    errors: list[str] = dataclasses.field(init=False, default_factory=list)

    def __post_init__(self) -> None:
        self._check_for_validate_functions(self.expected)

    def _str_errors(self) -> str:
        return "- " + "\n- ".join(self.errors)

    def _adderr(self, msg: str, *args, e=None) -> None:
        if e:
            logger.exception(msg, *args)
        else:
            logger.error(msg, *args)
        self.errors += [(msg % args)]

    @abstractmethod
    def verify(self, response):
        """Verify response against expected values and returns any values that
        we wanted to save for use in future requests

        It is expected that anything subclassing this can throw an exception indicating that the response
        verification failed.
        """

    def recurse_check_key_match(
        self,
        expected_block: Optional[Mapping],
        block: Mapping,
        blockname: str,
        strict: StrictOption,
    ) -> None:
        """Valid returned data against expected data

        Todo:
            Optionally use a validation library too

        Args:
            expected_block: expected data
            block: actual data
            blockname: 'name' of this block (params, mqtt, etc) for error messages
            strict: strictness setting for this block
        """

        if expected_block is None:
            logger.debug("No expected %s to check against", blockname)
            return

        # This should be done _before_ it gets to this point - typically in get_expected_from_request from plugin
        # expected_block = format_keys(
        #     expected_block, self.test_block_config.variables
        # )

        if block is None:
            if not expected_block:
                logger.debug(
                    "No %s in response to check, but not erroring because expected was %s",
                    blockname,
                    expected_block,
                )
                return

            self._adderr(
                "expected %s in the %s, but there was no response %s",
                expected_block,
                blockname,
                blockname,
            )
            return

        if isinstance(block, Mapping):
            block = dict(block)

        logger.debug("expected = %s, actual = %s", expected_block, block)

        try:
            check_keys_match_recursive(expected_block, block, [], strict)
        except exceptions.KeyMismatchError as e:
            self._adderr(e.args[0], e=e)

    def _check_for_validate_functions(self, response_block: Mapping) -> None:
        """
        See if there were any functions specified in the response block and save them for later use

        Args:
            response_block: block of external functions to call
        """

        def check_ext_functions(verify_block):
            if isinstance(verify_block, list):
                for vf in verify_block:
                    self.validate_functions.append(get_wrapped_response_function(vf))
            elif isinstance(verify_block, dict):
                self.validate_functions.append(
                    get_wrapped_response_function(verify_block)
                )
            elif verify_block is not None:
                raise exceptions.BadSchemaError(
                    "Badly formatted 'verify_response_with' block"
                )

        if mqtt_responses := response_block.get("mqtt_responses"):
            for mqtt_response in mqtt_responses:
                check_ext_functions(mqtt_response.get("verify_response_with", None))
        else:
            check_ext_functions(response_block.get("verify_response_with", None))

        def check_deprecated_validate(name):
            nfuncs = len(self.validate_functions)
            block = response_block.get(name, {})
            if isinstance(block, dict):
                check_ext_functions(block.get("$ext", None))
                if nfuncs != len(self.validate_functions):
                    raise exceptions.MisplacedExtBlockException(
                        name,
                    )

        # Could put in an isinstance check here
        check_deprecated_validate("json")

    def _maybe_run_validate_functions(self, response: Any) -> None:
        """Run validation functions if available

        Note:
             'response' will be different depending on where this is called from

        Args:
            response: Response type. This could be whatever the response type/plugin uses.
        """
        logger.debug(
            "Calling ext function from '%s' with response '%s'", type(self), response
        )

        for vf in self.validate_functions:
            try:
                vf(response)
            except Exception as e:
                self._adderr(
                    "Error calling validate function '%s':\n%s",
                    vf.func,
                    indent_err_text(traceback.format_exc()),
                    e=e,
                )

    def maybe_get_save_values_from_ext(
        self, response: Any, read_save_from: Mapping
    ) -> Mapping:
        """If there is an $ext function in the save block, call it and save the response

        Args:
            response: response object. Actual contents depends on which type of
                response is being checked
            read_save_from: the expected response (incl
                body/json/headers/mqtt topic/etc etc) containing a spec for which things
                should be saved from the response. Actual contents depends on which type of
                response is being checked

        Returns:
            mapping of name to value of things to save
        """

        try:
            wrapped = get_wrapped_response_function(read_save_from["save"]["$ext"])
        except KeyError:
            logger.debug("No save function for this stage")
            return {}

        try:
            saved = wrapped(response)
        except Exception as e:
            self._adderr(
                "Error calling save function '%s':\n%s",
                wrapped.func,  # type:ignore
                indent_err_text(traceback.format_exc()),
                e=e,
            )
            return {}

        logger.debug("saved %s from ext function", saved)

        if isinstance(saved, dict):
            return saved
        elif saved is not None:
            self._adderr(
                "Unexpected return value '%s' from $ext save function (expected a dict or None)",
                saved,
            )

        return {}

    def maybe_get_save_values_from_save_block(
        self,
        key: str,
        save_from: Optional[Mapping],
        *,
        outer_save_block: Optional[Mapping] = None,
    ) -> Mapping:
        """Save a value from a specific block in the response.

        See docs for maybe_get_save_values_from_given_block for more info

        Args:
            key: Name of key being used to save, for debugging
            save_from: An element of the response from which values are being saved
            outer_save_block: Read things to save from this block instead of self.expected
        """

        logger.debug("save from: %s", save_from)

        read_save_from = outer_save_block or self.expected
        logger.debug("save spec: %s", read_save_from.get("save"))

        try:
            to_save = read_save_from["save"][key]
        except KeyError:
            logger.debug("Nothing expected to save for %s", key)
            return {}

        return self.maybe_get_save_values_from_given_block(key, save_from, to_save)

    def maybe_get_save_values_from_given_block(
        self,
        key: str,
        save_from: Optional[Mapping],
        to_save: Mapping,
    ) -> Mapping:
        """Save a value from a specific block in the response.

        This is different from maybe_get_save_values_from_ext - depends on the kind of response

        Args:
            key: Name of key being used to save, for debugging
            save_from: An element of the response from which values are being saved
            to_save: block containing information about things to save

        Returns:
            mapping of save_name: value, where save_name is the key we
                wanted to save this value as
        """

        saved = {}

        if not save_from:
            self._adderr("No %s in response (wanted to save %s)", key, to_save)
            return {}

        for save_as, joined_key in to_save.items():
            try:
                saved[save_as] = recurse_access_key(save_from, joined_key)
            except (
                exceptions.InvalidQueryResultTypeError,
                exceptions.KeySearchNotFoundError,
            ) as e:
                self._adderr(
                    "Wanted to save '%s' from '%s', but it did not exist in the response",
                    joined_key,
                    key,
                    e=e,
                )

        if saved:
            logger.debug("Saved %s for '%s' from response", saved, key)

        return saved



import logging
from abc import abstractmethod
from typing import Any

import box

from tavern._core.pytest.config import TestConfig

logger: logging.Logger = logging.getLogger(__name__)


class BaseRequest:
    @abstractmethod
    def __init__(
        self, session: Any, rspec: dict, test_block_config: TestConfig
    ) -> None: ...

    @property
    @abstractmethod
    def request_vars(self) -> box.Box:
        """
        Variables used in the request

        What is contained in the return value will change depending on the type of request

        Returns:
            box.Box: box of request vars
        """

    @abstractmethod
    def run(self):
        """Run test"""



import contextlib
import copy
import logging
import os
import tempfile
from collections.abc import Mapping

import pykwalify
import yaml
from pykwalify import core

from tavern._core.exceptions import BadSchemaError
from tavern._core.loader import load_single_document_yaml
from tavern._core.plugins import load_plugins
from tavern._core.schema.jsonschema import verify_jsonschema

logger: logging.Logger = logging.getLogger(__name__)


class SchemaCache:
    """Caches loaded schemas"""

    def __init__(self) -> None:
        self._loaded: dict[str, dict] = {}

    def _load_base_schema(self, schema_filename):
        try:
            return self._loaded[schema_filename]
        except KeyError:
            self._loaded[schema_filename] = load_single_document_yaml(schema_filename)

            logger.debug("Loaded schema from %s", schema_filename)

            return self._loaded[schema_filename]

    def _load_schema_with_plugins(self, schema_filename: str) -> dict:
        mangled = f"{schema_filename}-plugins"

        try:
            return self._loaded[mangled]
        except KeyError:
            pass

        plugins = load_plugins()
        base_schema = copy.deepcopy(self._load_base_schema(schema_filename))

        logger.debug("Adding plugins to schema: %s", [p.name for p in plugins])

        for p in plugins:
            try:
                plugin_schema = p.plugin.schema
            except AttributeError:
                # Don't require a schema
                logger.debug("No schema defined for %s", p.name)
            else:
                base_schema["properties"].update(plugin_schema.get("properties", {}))

        self._loaded[mangled] = base_schema
        return self._loaded[mangled]

    def __call__(self, schema_filename: str, with_plugins: bool):
        """Load the schema file and cache it for future use

        Args:
            schema_filename: filename of schema
            with_plugins: Whether to load plugin schema into this schema as well

        Returns:
            loaded schema
        """

        if with_plugins:
            schema = self._load_schema_with_plugins(schema_filename)
        else:
            schema = self._load_base_schema(schema_filename)

        return schema


load_schema_file = SchemaCache()


def verify_pykwalify(to_verify, schema) -> None:
    """Verify a generic file against a given pykwalify schema
    Args:
        to_verify: Filename of source tests to check
        schema: Schema to verify against
    Raises:
        BadSchemaError: Schema did not match
    """
    logger.debug("Verifying %s against %s", to_verify, schema)

    here = os.path.dirname(os.path.abspath(__file__))
    extension_module_filename = os.path.join(here, "extensions.py")

    verifier = core.Core(
        source_data=to_verify,
        schema_data=schema,
        extensions=[extension_module_filename],
    )

    try:
        verifier.validate()
    except pykwalify.errors.PyKwalifyException as e:
        logger.exception("Error validating %s", to_verify)
        raise BadSchemaError() from e


@contextlib.contextmanager
def wrapfile(to_wrap):
    """Wrap a dictionary into a temporary yaml file

    Args:
        to_wrap: Dictionary to write to temporary file

    Yields:
        filename: name of temporary file object that will be destroyed at the end of the
            context manager
    """
    with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as wrapped_tmp:
        # put into a file
        dumped = yaml.dump(to_wrap, default_flow_style=False)
        wrapped_tmp.write(dumped.encode("utf8"))
        wrapped_tmp.close()

        try:
            yield wrapped_tmp.name
        finally:
            os.remove(wrapped_tmp.name)


def verify_tests(test_spec: Mapping, with_plugins: bool = True) -> None:
    """Verify that a specific test block is correct

    Todo:
        Load schema file once. Requires some caching of the file

    Args:
        test_spec: Test in dictionary form
        with_plugins: Whether to load plugin schema into this schema as well

    Raises:
        BadSchemaError: Schema did not match
    """
    here = os.path.dirname(os.path.abspath(__file__))

    schema_filename = os.path.join(here, "tests.jsonschema.yaml")
    schema = load_schema_file(schema_filename, with_plugins)

    verify_jsonschema(test_spec, schema)



import dataclasses
import logging
from collections.abc import Iterable, Mapping
from typing import (
    Optional,
    Protocol,
    Union,
)

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass
class YamlMark:
    """A pyyaml mark"""

    line: int = 0
    name: Optional[str] = None


class _WithMarks(Protocol):
    """Things loaded by pyyaml have these"""

    start_mark: YamlMark
    end_mark: YamlMark


PyYamlDict = Union[_WithMarks, Mapping]


def get_stage_lines(stage: PyYamlDict) -> tuple[int, int, int]:
    first_line = start_mark(stage).line - 1
    last_line = end_mark(stage).line
    line_start = first_line + 1

    return first_line, last_line, line_start


def read_relevant_lines(
    yaml_block: PyYamlDict, first_line: int, last_line: int
) -> Iterable[str]:
    """Get lines between start and end mark"""

    filename = get_stage_filename(yaml_block)

    if filename is None:
        logger.warning("unable to read yaml block")
        return

    with open(filename, encoding="utf8") as testfile:
        for idx, line in enumerate(testfile.readlines()):
            if first_line < idx < last_line:
                yield line.split("#", 1)[0].rstrip()


def get_stage_filename(yaml_block: PyYamlDict) -> Optional[str]:
    return start_mark(yaml_block).name


def start_mark(yaml_block: PyYamlDict) -> Union[type[YamlMark], YamlMark]:
    try:
        return yaml_block.start_mark  # type:ignore
    except AttributeError:
        return YamlMark()


def end_mark(yaml_block: PyYamlDict) -> Union[type[YamlMark], YamlMark]:
    try:
        return yaml_block.end_mark  # type:ignore
    except AttributeError:
        return YamlMark()



from unittest.mock import Mock, patch

import pytest

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.loader import ANYTHING
from tavern._plugins.rest.response import RestResponse


@pytest.fixture(name="example_response")
def fix_example_response():
    spec = {
        "status_code": 302,
        "headers": {
            "Content-Type": "application/json",
            "location": "www.google.com?search=breadsticks",
        },
        "json": {"a_thing": "authorization_code", "code": "abc123"},
    }

    return spec.copy()


@pytest.fixture(name="nested_response")
def fix_nested_response():
    # https://github.com/taverntesting/tavern/issues/45
    spec = {
        "status_code": 200,
        "headers": {"Content-Type": "application/json"},
        "json": {"users": [{"u": {"user_id": "def456"}}]},
    }

    return spec.copy()


@pytest.fixture(name="nested_schema")
def fix_nested_schema():
    # https://github.com/taverntesting/tavern/issues/45
    spec = {
        "status_code": 200,
        "headers": {"Content-Type": "application/json"},
        "json": {"users": [{"u": {"user_id": "{code}"}}]},
    }

    return spec.copy()


class TestSave:
    def test_save_body(self, example_response, includes):
        """Save a key from the body into the right name"""
        example_response["save"] = {"json": {"test_code": "code"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(
            "json", example_response["json"]
        )

        assert saved == {"test_code": example_response["json"]["code"]}

    def test_save_body_nested(self, example_response, includes):
        """Save a key from the body into the right name"""
        example_response["json"]["nested"] = {"subthing": "blah"}
        example_response["save"] = {"json": {"test_nested_thing": "nested.subthing"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(
            "json", example_response["json"]
        )

        assert saved == {
            "test_nested_thing": example_response["json"]["nested"]["subthing"]
        }

    def test_save_body_nested_list(self, example_response, includes):
        """Save a key from the body into the right name"""
        example_response["json"]["nested"] = {"subthing": ["abc", "def"]}
        example_response["save"] = {"json": {"test_nested_thing": "nested.subthing[0]"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(
            "json", example_response["json"]
        )

        assert saved == {
            "test_nested_thing": example_response["json"]["nested"]["subthing"][0]
        }

    def test_save_header(self, example_response, includes):
        """Save a key from the headers into the right name"""
        example_response["save"] = {"headers": {"next_location": "location"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(
            "headers", example_response["headers"]
        )

        assert saved == {"next_location": example_response["headers"]["location"]}

    def test_save_redirect_query_param(self, example_response, includes):
        """Save a key from the query parameters of the redirect location"""
        example_response["save"] = {"redirect_query_params": {"test_search": "search"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(
            "redirect_query_params", {"search": "breadsticks"}
        )

        assert saved == {"test_search": "breadsticks"}

    @pytest.mark.parametrize("save_from", ("json", "headers", "redirect_query_params"))
    def test_bad_save(self, save_from, example_response, includes):
        example_response["save"] = {save_from: {"abc": "123"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        saved = r.maybe_get_save_values_from_save_block(save_from, {})

        assert not saved

        assert r.errors


class TestValidate:
    def test_simple_validate_body(self, example_response, includes):
        """Make sure a simple value comparison works"""

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", example_response["json"])

        assert not r.errors

    def test_validate_list_body(self, example_response, includes):
        """Make sure a list response can be validated"""

        example_response["json"] = ["a", 1, "b"]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", example_response["json"])

        assert not r.errors

    def test_validate_list_body_wrong_order(self, example_response, includes):
        """Order of list items matters"""

        example_response["json"] = ["a", 1, "b"]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", example_response["json"][::-1])

        assert r.errors

    def test_validate_nested_body(self, example_response, includes):
        """Make sure a nested value comparison works"""

        example_response["json"]["nested"] = {"subthing": "blah"}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", example_response["json"])

        assert not r.errors

    def test_simple_validate_headers(self, example_response, includes):
        """Make sure a simple value comparison works"""

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("headers", example_response["headers"])

        assert not r.errors

    def test_simple_validate_redirect_query_params(self, example_response, includes):
        """Make sure a simple value comparison works"""

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("redirect_query_params", {"search": "breadsticks"})

        assert not r.errors

    def test_validate_missing_list_key(self, example_response, includes):
        """If we expect 4 items and 3 were returned, catch error"""

        example_response["json"] = ["a", 1, "b", "c"]
        bad_expected = example_response["json"][:-1]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", bad_expected)

        assert r.errors

    def test_validate_wrong_list_dict(self, example_response, includes):
        """We expected a list, but we got a dict in the response"""

        example_response["json"] = ["a", 1, "b", "c"]
        bad_expected = {"a": "b"}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", bad_expected)

        assert r.errors

    def test_validate_wrong_dict_list(self, example_response, includes):
        """We expected a dict, but we got a list in the response"""

        example_response["json"] = {"a": "b"}
        bad_expected = ["a", "b", "c"]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", bad_expected)

        assert r.errors


class TestMatchStatusCodes:
    def test_validate_single_status_code_passes(self, example_response, includes):
        """single status code match"""

        example_response["status_code"] = 100

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._check_status_code(100, {})

        assert not r.errors

    def test_validate_single_status_code_incorrect(self, example_response, includes):
        """single status code mismatch"""

        example_response["status_code"] = 100

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._check_status_code(102, {})

        assert r.errors

    def test_validate_multiple_status_codes_passes(self, example_response, includes):
        """Check it can match mutliple status codes"""

        example_response["status_code"] = [100, 200, 300]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._check_status_code(100, {})

        assert not r.errors

    def test_validate_multiple_status_codes_missing(self, example_response, includes):
        """Status code was not in list"""

        example_response["status_code"] = [100, 200, 300]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._check_status_code(103, {})

        assert r.errors


class TestNestedValidate:
    def test_validate_nested_null(self, example_response, includes):
        """Check that nested 'null' comparisons do not work"""

        example_response["json"] = {"nested": {"subthing": None}}

        expected = {"nested": {"subthing": "blah"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", expected)

        assert r.errors

    def test_validate_nested_anything(self, example_response, includes):
        """Check that nested 'anything' comparisons work

        This is a bit hacky because we're directly checking the ANYTHING
        comparison - need to add an integration test too
        """

        example_response["json"] = {"nested": {"subthing": ANYTHING}}

        expected = {"nested": {"subthing": "blah"}}

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        r._validate_block("json", expected)

        assert not r.errors


class TestFull:
    def test_validate_and_save(self, example_response, includes):
        """Test full verification + return saved values"""
        example_response["save"] = {"json": {"test_code": "code"}}
        r = RestResponse(Mock(), "Test 1", example_response, includes)

        class FakeResponse:
            headers = example_response["headers"]
            content = b"test"

            def json(self):
                return example_response["json"]

            status_code = example_response["status_code"]

        saved = r.verify(FakeResponse())

        assert saved == {"test_code": example_response["json"]["code"]}

    def test_incorrect_status_code(self, example_response, includes):
        """Test full verification + return saved values"""
        r = RestResponse(Mock(), "Test 1", example_response, includes)

        class FakeResponse:
            headers = example_response["headers"]
            content = b"test"

            def json(self):
                return example_response["json"]

            status_code = 400

        with pytest.raises(exceptions.TestFailError):
            r.verify(FakeResponse())

        assert r.errors

    def test_saved_value_in_validate(self, nested_response, nested_schema, includes):
        r = RestResponse(
            Mock(),
            "Test 1",
            format_keys(nested_schema, includes.variables),
            includes,
        )

        class FakeResponse:
            headers = nested_response["headers"]
            content = b"test"

            def json(self):
                return nested_response["json"]

            status_code = nested_response["status_code"]

        r.verify(FakeResponse())

    @pytest.mark.parametrize("value", [1, "some", False, None])
    def test_validate_single_value_response(self, example_response, includes, value):
        """Check validating single value response (string, int, etc)."""
        del example_response["json"]

        r = RestResponse(Mock(), "Test 1", example_response, includes)

        class FakeResponse:
            headers = example_response["headers"]
            content = b"test"

            def json(self):
                return value

            status_code = example_response["status_code"]

        r.verify(FakeResponse())


def test_status_code_warns(example_response, includes):
    """Should continue if the status code is nonexistent"""
    example_response["status_code"] = 231234

    with patch("tavern._plugins.rest.response.logger.warning") as wmock:
        RestResponse(Mock(), "Test 1", example_response, includes)

    assert wmock.called



import logging
from collections.abc import MutableMapping
from typing import Any

from tavern._core.pytest.config import TestConfig

logger: logging.Logger = logging.getLogger(__name__)


def pytest_tavern_beta_before_every_test_run(test_dict: dict, variables: dict) -> None:
    """Called:

    - directly after fixtures are loaded for a test
    - directly before verifying the schema of the file
    - Before formatting is done on values
    - After global configuration has been loaded
    - After plugins have been loaded

    Modify the test in-place if you want to do something to it.

    Args:
        test_dict: Test to run
        variables: Available variables
    """


def pytest_tavern_beta_after_every_test_run(test_dict: dict, variables: dict) -> None:
    """Called:

    - After test run

    Args:
        test_dict: Test to run
        variables: Available variables
    """


def pytest_tavern_beta_after_every_response(expected: Any, response: Any) -> None:
    """Called after every _response_ - including MQTT/HTTP/etc

    Note:
        - The response object type and the expected dict depends on what plugin you're using, and which kind of response it is!
        - MQTT responses will call this hook multiple times if multiple messages are received

    Args:
        expected: Response block in stage
        response: Response object.
    """


def pytest_tavern_beta_before_every_request(request_args: MutableMapping) -> None:
    """Called just before every request - including MQTT/HTTP/etc

    Note:
        - The request object type depends on what plugin you're using, and which kind of request it is!

    Args:
        request_args: Arguments passed to the request function. By default, this is Session.request for
            HTTP and Client.publish for MQTT
    """


def call_hook(test_block_config: TestConfig, hookname: str, **kwargs) -> None:
    """Utility to call the hooks"""
    try:
        hook = getattr(test_block_config.tavern_internal.pytest_hook_caller, hookname)
    except AttributeError:
        logger.critical("Error getting tavern hook!")
        raise

    try:
        hook(**kwargs)
    except AttributeError:
        logger.error("Unexpected error calling tavern hook")
        raise



import dataclasses
import json
import logging
import re
import typing
from io import StringIO
from typing import Any, Optional

import yaml
from _pytest._code.code import FormattedExcinfo, TerminalRepr
from _pytest._io import TerminalWriter

from tavern._core import exceptions
from tavern._core.dict_util import format_keys

if typing.TYPE_CHECKING:
    from tavern._core.pytest.item import YamlItem

from tavern._core.report import prepare_yaml
from tavern._core.stage_lines import (
    end_mark,
    get_stage_lines,
    read_relevant_lines,
    start_mark,
)

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ReprdError(TerminalRepr):
    exce: Any
    item: "YamlItem"

    def _get_available_format_keys(self) -> dict:
        """Try to get the format variables for the stage

        If we can't get the variable for this specific stage, just return the
        global config which will at least have some format variables

        Returns:
            variables for formatting test
        """
        try:
            keys = self.exce._excinfo[1].test_block_config.variables
        except AttributeError:
            logger.warning("Unable to read stage variables - error output may be wrong")
            keys = self.item.global_cfg.variables

        return keys

    def _print_format_variables(
        self, tw: TerminalWriter, code_lines: list[str]
    ) -> list[str]:
        """Print a list of the format variables and their value at this stage

        If the format variable is not defined, print it in red as '???'

        Args:
            tw: Pytest TW instance
            code_lines: Source lines for this stage

        Returns:
            List of all missing format variables
        """

        def read_formatted_vars(lines):
            """Go over all lines and try to find format variables"""
            for line in lines:
                for match in re.finditer(
                    r"(.*?:\s+!raw)?(?(1).*|.*?(?P<format_var>(?<!{){[^{]*?}))", line
                ):
                    if match.group("format_var") is not None:
                        yield match.group("format_var")

        format_variables = set(read_formatted_vars(code_lines))

        keys = self._get_available_format_keys()

        missing = []

        # Print out values of format variables, like Pytest prints out the
        # values of function call variables
        tw.line("Format variables:", white=True, bold=True)
        for var in format_variables:
            # Empty dict
            if re.match(r"^\s*{}\s*", var):
                continue

            # If it's valid json, it's not a format value
            try:
                json.loads(var)
            except json.JSONDecodeError:
                pass
            else:
                continue

            try:
                value_at_call = format_keys(var, keys)
            except exceptions.MissingFormatError:
                missing.append(var)
                value_at_call = "???"
                white = False
                red = True
            else:
                white = True
                red = False

            line = f"  {var[1:-1]} = '{value_at_call}'"
            tw.line(line, white=white, red=red)  # pragma: no cover

        return missing

    def _print_test_stage(
        self,
        tw: TerminalWriter,
        code_lines: list[str],
        missing_format_vars: list[str],
        line_start: Optional[int],
    ) -> None:
        """Print the direct source lines from this test stage

        If we couldn't get the stage for some reason, print the entire test out.

        If there are any lines which have missing format variables, higlight
        them in red.

        Args:
            tw: Pytest TW instance
            code_lines: Raw source for this stage
            missing_format_vars: List of all missing format
                variables for this stage
            line_start: Source line of this stage
        """
        if line_start:
            tw.line(f"Source test stage (line {line_start}):", white=True, bold=True)
        else:
            tw.line("Source test stages:", white=True, bold=True)

        for line in code_lines:
            if any(i in line for i in missing_format_vars):
                tw.line(line, red=True)
            else:
                tw.line(line, white=True)

    def _print_formatted_stage(self, tw: TerminalWriter, stage: dict) -> None:
        """Print the 'formatted' stage that Tavern will actually use to send the
        request/process the response

        Args:
            tw: Pytest TW instance
            stage: The 'final' stage used by Tavern
        """
        tw.line("Formatted stage:", white=True, bold=True)

        keys = self._get_available_format_keys()

        # Format stage variables recursively
        formatted_stage = format_keys(
            stage, keys, dangerously_ignore_string_format_errors=True
        )

        # Replace formatted strings with strings for dumping
        prepared_stage = prepare_yaml(formatted_stage)

        # Dump formatted stage to YAML format
        formatted_lines = yaml.dump(prepared_stage, default_flow_style=False).split(
            "\n"
        )

        for line in formatted_lines:
            if not line:
                continue
            tw.line(f"  {line}", white=True)

    def _print_errors(self, tw: TerminalWriter) -> None:
        """Print any errors in the 'normal' Pytest style

        Args:
            tw: Pytest TW instance
        """
        tw.line("Errors:", white=True, bold=True)

        # Sort of hack, just use this to directly extract the exception format.
        # If this breaks in future, just re-implement it
        e = FormattedExcinfo()
        lines = e.get_exconly(self.exce)
        for line in lines:
            tw.line(line, red=True, bold=True)

    def toterminal(self, tw: TerminalWriter) -> None:
        """Print out a custom error message to the terminal"""

        # Try to get the stage so we can print it out. I'm not sure if the stage
        # will ever NOT be present, but better to check just in case
        try:
            stage = self.exce._excinfo[1].stage
        except AttributeError:
            stage = None
            # Fallback, we don't know which stage it is
            stages = self.item.spec["stages"]

            first_line = start_mark(stages[0]).line - 1
            last_line = end_mark(stages[-1]).line

            line_start = None
        else:
            first_line, last_line, line_start = get_stage_lines(stage)

        code_lines = list(read_relevant_lines(self.item.spec, first_line, last_line))

        missing_format_vars = self._print_format_variables(tw, code_lines)
        tw.line("")

        self._print_test_stage(tw, code_lines, missing_format_vars, line_start)
        tw.line("")

        if not stage:
            tw.line(
                "[Could not determine which stage was running]", red=True, bold=True
            )
        elif missing_format_vars:
            tw.line("Missing format vars for stage", red=True, bold=True)
        else:
            self._print_formatted_stage(tw, stage)

        tw.line("")

        self._print_errors(tw)

    @property
    def longreprtext(self) -> str:
        # information.
        io = StringIO()
        tw = TerminalWriter(file=io)
        self.toterminal(tw)
        return io.getvalue().strip()

    def __str__(self) -> str:
        return self.longreprtext



import dataclasses
import logging
import unittest.mock
from collections.abc import Mapping
from unittest.mock import patch

import pytest

from tavern._core import exceptions
from tavern._core.pytest.config import TestConfig
from tavern._core.run import run_test


def _run_test(
    stage: Mapping, test_block_config: TestConfig, run_mock: unittest.mock.Mock
) -> bool:
    """runs the test and returns whether the stage was run or not"""

    full_test = {
        "test_name": "A test with a single stage",
        "stages": [stage],
    }

    run_test("test_file_name", full_test, test_block_config)

    return run_mock.called


class TestSkipStage:
    @pytest.fixture(autouse=True)
    def run_mock(self):
        with patch("tavern._core.run._TestRunner.run_stage") as run_mock:
            yield run_mock

    @pytest.fixture(scope="function")
    def stage(self):
        return {
            "name": "test stage",
            "request": {"url": "https://example.com", "method": "GET"},
            "response": {"status_code": 200},
        }

    @pytest.fixture
    def test_block_config(self, includes):
        return dataclasses.replace(includes, variables={"env_vars": {}})

    def test_skip_true(self, stage, test_block_config, run_mock):
        """Skip stage when 'skip' is True"""

        stage["skip"] = True
        assert _run_test(stage, test_block_config, run_mock) is False

    def test_skip_false(self, stage, test_block_config, run_mock):
        """Don't skip stage when 'skip' is False"""

        stage["skip"] = False
        assert _run_test(stage, test_block_config, run_mock) is True

    def test_skip_simpleeval_true(self, stage, test_block_config, run_mock):
        """Skip stage when simpleeval expression evaluates to True"""

        stage["skip"] = "True"
        assert _run_test(stage, test_block_config, run_mock) is False

    def test_skip_simpleeval_false(self, stage, test_block_config, run_mock):
        """Don't skip stage when simpleeval expression evaluates to False"""

        stage["skip"] = "False"
        assert _run_test(stage, test_block_config, run_mock) is True

    def test_skip_env_var_true(self, stage, test_block_config, run_mock):
        """Skip stage when using a variable that evaluates to True"""

        stage["skip"] = "'{some_var}' == 'value'"
        test_block_config.variables.update({"some_var": "value"})

        assert _run_test(stage, test_block_config, run_mock) is False

    def test_skip_env_var_false(self, stage, test_block_config, run_mock):
        """Don't skip stage when using a variable that evaluates to False"""

        stage["skip"] = "'{some_var}' == 'value'"
        test_block_config.variables.update({"some_var": "value"})

        assert _run_test(stage, test_block_config, run_mock) is False

    def test_skip_invalid_var_types(self, stage, test_block_config, run_mock):
        """Error when cel types are wrong"""

        stage["skip"] = "'{some_var}' > 3"
        test_block_config.variables.update({"some_var": "value"})

        with pytest.raises(exceptions.EvalError):
            _run_test(stage, test_block_config, run_mock)

    @pytest.mark.xfail(
        reason="'KeyError: <_pytest.stash.StashKey object at 0x7fa6ac4852c0' ?????"
    )
    def test_skip_invalid_simpleeval(self, stage, test_block_config, caplog, run_mock):
        """Handle invalid simpleeval expressions gracefully"""

        stage["skip"] = "hello i am a test <<<"

        with caplog.at_level(logging.WARNING):
            _run_test(stage, test_block_config, run_mock)

        assert "unable to parse as simpleeval" in caplog.text

    def test_error_valid_simpleeval_missing_var(
        self, stage, test_block_config, run_mock
    ):
        """Handle missing variable"""

        stage["skip"] = "invalid_cel_expression"

        with pytest.raises(exceptions.EvalError):
            _run_test(stage, test_block_config, run_mock)

    def test_skip_non_bool_result(self, stage, test_block_config, run_mock):
        """Raise error when CEL returns non-boolean value"""

        stage["skip"] = "'not a boolean'"
        with pytest.raises(exceptions.EvalError):
            _run_test(stage, test_block_config, run_mock)

    def test_skip_empty_string(self, stage, test_block_config, run_mock):
        """Treat empty string as False"""

        stage["skip"] = ""
        assert _run_test(stage, test_block_config, run_mock) is True



import contextlib
import datetime
import functools
import sqlite3

import jwt
from flask import Flask, g, jsonify, request

app = Flask(__name__)


SECRET = "CGQgaG7GYvTcpaQZqosLy4"
DATABASE = "/tmp/test_db"
SERVERNAME = "testserver"


def get_db():
    db = getattr(g, "_database", None)
    if db is None:
        db = g._database = sqlite3.connect(DATABASE)

        with db:
            with contextlib.suppress(Exception):
                db.execute(
                    "CREATE TABLE numbers_table (name TEXT NOT NULL, number INTEGER NOT NULL)"
                )

    return db


@app.teardown_appcontext
def close_connection(exception):
    db = getattr(g, "_database", None)
    if db is not None:
        db.close()


@app.route("/login", methods=["POST"])
def login():
    r = request.get_json()

    if r["user"] != "test-user" or r["password"] != "correct-password":
        return jsonify({"error": "Incorrect username/password"}), 401

    payload = {
        "sub": "test-user",
        "aud": SERVERNAME,
        "exp": datetime.datetime.utcnow() + datetime.timedelta(hours=1),
    }

    token = jwt.encode(payload, SECRET, algorithm="HS256")

    return jsonify({"token": token})


def requires_jwt(endpoint):
    """Makes sure a jwt is in the request before accepting it"""

    @functools.wraps(endpoint)
    def check_auth_call(*args, **kwargs):
        token = request.headers.get("Authorization")

        # check token is present
        if not token:
            return jsonify({"error": "No token"}), 401

        token_type, token = token.split(" ")

        if token_type.lower() != "bearer":
            return jsonify({"error": "Wrong token type"}), 401

        try:
            jwt.decode(token, SECRET, audience=SERVERNAME, algorithms=["HS256"])
        except Exception:
            return jsonify({"error": "Invalid token"}), 401

        return endpoint(*args, **kwargs)

    return check_auth_call


@app.route("/numbers", methods=["POST"])
@requires_jwt
def add_number():
    r = request.get_json()

    try:
        r["number"]
        r["name"]
    except (KeyError, TypeError):
        return jsonify({"error": "missing key"}), 400

    db = get_db()

    with db:
        db.execute("INSERT INTO numbers_table VALUES (:name, :number)", r)

    return jsonify({}), 201


@app.route("/numbers", methods=["GET"])
@requires_jwt
def get_number():
    r = request.args

    try:
        r["name"]
    except (KeyError, TypeError):
        return jsonify({"error": "missing key"}), 400

    db = get_db()

    with db:
        row = db.execute("SELECT number FROM numbers_table WHERE name IS :name", r)

    try:
        number = next(row)[0]
    except StopIteration:
        return jsonify({"error": "Unknown number"}), 404

    return jsonify({"number": number})


@app.route("/double", methods=["POST"])
@requires_jwt
def double_number():
    r = request.get_json()

    try:
        r["name"]
    except (KeyError, TypeError):
        return jsonify({"error": "no number passed"}), 400

    db = get_db()

    with db:
        db.execute("UPDATE numbers_table SET number = number*2 WHERE name IS :name", r)
        row = db.execute("SELECT number FROM numbers_table WHERE name IS :name", r)

    try:
        double = next(row)[0]
    except StopIteration:
        return jsonify({"error": "Unknown number"}), 404

    return jsonify({"number": double})


@app.route("/reset", methods=["POST"])
def reset_db():
    db = get_db()

    with db:
        db.execute("DELETE FROM numbers_table")

    return "", 204



import random
import re
import threading
from unittest.mock import Mock, patch

import pytest

from tavern._core import exceptions
from tavern._core.loader import ANYTHING
from tavern._core.strict_util import StrictLevel
from tavern._plugins.mqtt.client import MQTTClient
from tavern._plugins.mqtt.response import MQTTResponse


def test_nothing_returned_fails(includes):
    """Raises an error if no message was received"""
    fake_client = Mock(spec=MQTTClient, message_received=Mock(return_value=None))

    expected = {"mqtt_responses": [{"topic": "/a/b/c", "payload": "hello"}]}

    verifier = MQTTResponse(
        fake_client,
        "Test stage",
        expected,
        includes.with_strictness(StrictLevel.all_on()),
    )

    with pytest.raises(exceptions.TestFailError):
        verifier.verify(expected)

    assert not verifier.received_messages


class FakeMessage:
    def __init__(self, returned):
        self.topic = returned["topic"]
        self.payload = returned["payload"].encode("utf8")
        self.timestamp = 0


class TestResponse:
    @staticmethod
    def _get_fake_verifier(expected, fake_messages, includes):
        """Given a list of messages, return a mocked version of the MQTT
        response verifier which will take messages off the front of this list as
        if they were published

        This mocks it as if all messages were returned in order, which they
        might not have been...?
        """
        if not isinstance(fake_messages, list):
            pytest.fail("Need to pass a list of messages")

        msg_lock = threading.RLock()

        responses: dict[str, list[FakeMessage]] = {
            message.topic: [] for message in fake_messages
        }
        for message in fake_messages:
            responses[message.topic].append(message)

        def yield_all_messages():
            def inner(topic, timeout):
                try:
                    msg_lock.acquire()

                    r = responses[topic]
                    if len(r) == 0:
                        return None

                    return r.pop(random.randint(0, len(r) - 1))
                finally:
                    msg_lock.release()

            return inner

        fake_client = Mock(
            spec=MQTTClient,
            message_received=yield_all_messages(),
        )

        if not isinstance(expected, list):
            expected = [expected]

        return MQTTResponse(
            fake_client, "Test stage", {"mqtt_responses": expected}, includes
        )

    def test_message_on_same_topic_fails(self, includes):
        """Correct topic, wrong message"""

        expected = {"topic": "/a/b/c", "payload": "hello"}

        fake_message = FakeMessage({"topic": "/a/b/c", "payload": "goodbye"})

        verifier = self._get_fake_verifier(expected, [fake_message], includes)

        with pytest.raises(exceptions.TestFailError):
            verifier.verify(expected)

        assert len(verifier.received_messages) == 1
        assert verifier.received_messages[0].topic == fake_message.topic

    def test_correct_message(self, includes):
        """Both correct matches"""

        expected = {"topic": "/a/b/c", "payload": "hello"}

        fake_message = FakeMessage(expected)

        verifier = self._get_fake_verifier(expected, [fake_message], includes)

        verifier.verify(expected)

        assert len(verifier.received_messages) == 1
        assert verifier.received_messages[0].topic == fake_message.topic

    @pytest.mark.parametrize("n_messages", (1, 2))
    def test_ext_function_called_save(self, includes, n_messages: int):
        """Make sure that it calls ext functions appropriately on individual MQTT
        responses and saved the response"""
        expecteds = []
        fake_messages = []
        for i in range(n_messages):
            expected = {
                "topic": f"/a/b/c/{i + 1}",
                "payload": "hello",
                "save": {
                    "$ext": {"function": f"function_name_{i + 1}"},
                },
            }

            fake_message = FakeMessage(expected)

            expecteds += [expected]
            fake_messages += [fake_message]

        verifier = self._get_fake_verifier(expecteds, fake_messages, includes)

        def fake_get_wrapped_response():
            def wrap(ext):
                def actual(response, *args, **kwargs):
                    match = re.match(r"function_name_(?P<idx>\d+)", ext["function"])
                    assert match
                    message_number = match.group("idx")
                    return {f"saved_topic_{message_number}": response.topic}

                return actual

            return wrap

        with patch(
            "tavern.response.get_wrapped_response_function",
            new_callable=fake_get_wrapped_response,
        ):
            saved = verifier.verify(None)

        assert len(verifier.received_messages) == n_messages

        for i in range(n_messages):
            assert verifier.received_messages[i].topic == fake_messages[i].topic

            assert len(saved) == n_messages

            assert saved[f"saved_topic_{i + 1}"] == expecteds[i]["topic"]

    def test_correct_message_eventually(self, includes):
        """One wrong messge, then the correct one"""

        expected = {"topic": "/a/b/c", "payload": "hello"}

        fake_message_good = FakeMessage(expected)
        fake_message_bad = FakeMessage({"topic": "/a/b/c", "payload": "goodbye"})

        verifier = self._get_fake_verifier(
            expected, [fake_message_bad, fake_message_good], includes
        )

        verifier.verify(expected)

        assert len(verifier.received_messages) >= 1
        received_topics = [m.topic for m in verifier.received_messages]
        assert fake_message_good.topic in received_topics

    def test_unexpected_fail(self, includes):
        """Messages marked unexpected fail test"""

        expected = {"topic": "/a/b/c", "payload": "hello", "unexpected": True}

        fake_message = FakeMessage(expected)

        verifier = self._get_fake_verifier(expected, [fake_message], includes)

        with pytest.raises(exceptions.TestFailError):
            verifier.verify(expected)

        assert len(verifier.received_messages) == 1
        assert verifier.received_messages[0].topic == fake_message.topic

    @pytest.mark.parametrize("r", range(10))
    def test_multiple_messages(self, includes, r):
        """One wrong message, two correct ones"""

        expected = [
            {"topic": "/a/b/c", "payload": "hello"},
            {"topic": "/d/e/f", "payload": "hellog"},
        ]

        fake_message_good_1 = FakeMessage(expected[0])
        fake_message_good_2 = FakeMessage(expected[1])
        fake_message_bad = FakeMessage({"topic": "/a/b/c", "payload": "goodbye"})

        messages = [fake_message_bad, fake_message_good_1, fake_message_good_2]
        random.shuffle(messages)

        verifier = self._get_fake_verifier(
            expected,
            messages,
            includes,
        )

        verifier.verify(expected)

        assert len(verifier.received_messages) >= 2
        received_topics = [m.topic for m in verifier.received_messages]
        assert fake_message_good_1.topic in received_topics
        assert fake_message_good_2.topic in received_topics

    @pytest.mark.parametrize("r", range(10))
    def test_different_order(self, includes, r):
        """Messages coming in a different order"""

        expected = [
            {"topic": "/a/b/c", "payload": "hello"},
            {"topic": "/d/e/f", "payload": "hellog"},
        ]

        fake_message_good_1 = FakeMessage(expected[0])
        fake_message_good_2 = FakeMessage(expected[1])

        messages = [fake_message_good_2, fake_message_good_1]
        random.shuffle(messages)

        verifier = self._get_fake_verifier(expected, messages, includes)

        verifier.verify(expected)

        assert len(verifier.received_messages) == 2
        received_topics = [m.topic for m in verifier.received_messages]
        assert fake_message_good_1.topic in received_topics
        assert fake_message_good_2.topic in received_topics

    # FIXME: Add tests for 'ext' functions are called in the right order

    @pytest.mark.parametrize(
        "payload",
        (
            (
                "!anything",
                ANYTHING,
            ),
            (
                "null",
                None,
            ),
            (
                "goog",
                "goog",
            ),
        ),
    )
    @pytest.mark.parametrize("r", range(10))
    def test_same_topic(self, includes, r, payload):
        """Messages coming in a different order"""

        expected = [
            {"topic": "/a/b/c", "payload": "hello"},
            {"topic": "/a/b/c", "payload": payload[0]},
        ]

        fake_message_good_1 = FakeMessage(expected[0])
        fake_message_good_2 = FakeMessage(expected[1])

        messages = [fake_message_good_2, fake_message_good_1]
        random.shuffle(messages)

        verifier = self._get_fake_verifier(expected, messages, includes)

        loaded = [
            {"topic": "/a/b/c", "payload": "hello"},
            {"topic": "/a/b/c", "payload": payload[1]},
        ]
        verifier.verify(loaded)

        assert len(verifier.received_messages) == 2
        received_topics = [m.topic for m in verifier.received_messages]
        assert fake_message_good_1.topic in received_topics
        assert fake_message_good_2.topic in received_topics



# https://gist.github.com/joshbode/569627ced3076931b02f
import dataclasses
import logging
import os.path
import re
import uuid
from abc import abstractmethod
from itertools import chain
from typing import Optional, Union

import pytest
import yaml
from _pytest.python_api import ApproxBase
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.nodes import Node, ScalarNode
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner

from tavern._core import exceptions
from tavern._core.exceptions import BadSchemaError
from tavern._core.strtobool import strtobool

logger: logging.Logger = logging.getLogger(__name__)


def makeuuid(loader, node) -> str:
    return str(uuid.uuid4())


class RememberComposer(Composer):
    """A composer that doesn't forget anchors across documents"""

    def compose_document(self) -> Optional[Node]:
        # Drop the DOCUMENT-START event.
        self.get_event()  # type:ignore

        # Compose the root node.
        node = self.compose_node(None, None)  # type:ignore

        # Drop the DOCUMENT-END event.
        self.get_event()  # type:ignore

        # If we don't drop the anchors here, then we can keep anchors across
        # documents.
        # self.anchors = {}

        return node


def create_node_class(cls):
    class node_class(cls):
        def __init__(self, x, start_mark, end_mark):
            cls.__init__(self, x)
            self.start_mark = start_mark
            self.end_mark = end_mark

        # def __new__(self, x, start_mark, end_mark):
        #     return cls.__new__(self, x)

    node_class.__name__ = "%s_node" % cls.__name__
    return node_class


dict_node = create_node_class(dict)
list_node = create_node_class(list)


class SourceMappingConstructor(SafeConstructor):
    # To support lazy loading, the original constructors first yield
    # an empty object, then fill them in when iterated. Due to
    # laziness we omit this behaviour (and will only do "deep
    # construction") by first exhausting iterators, then yielding
    # copies.
    def construct_yaml_map(self, node):
        (obj,) = SafeConstructor.construct_yaml_map(self, node)
        return dict_node(obj, node.start_mark, node.end_mark)

    def construct_yaml_seq(self, node):
        (obj,) = SafeConstructor.construct_yaml_seq(self, node)
        return list_node(obj, node.start_mark, node.end_mark)


SourceMappingConstructor.add_constructor(  # type: ignore
    "tag:yaml.org,2002:map", SourceMappingConstructor.construct_yaml_map
)

SourceMappingConstructor.add_constructor(  # type: ignore
    "tag:yaml.org,2002:seq", SourceMappingConstructor.construct_yaml_seq
)

yaml.add_representer(dict_node, yaml.representer.SafeRepresenter.represent_dict)
yaml.add_representer(list_node, yaml.representer.SafeRepresenter.represent_list)


class IncludeLoader(
    Reader,
    Scanner,
    Parser,
    RememberComposer,
    Resolver,
    SourceMappingConstructor,
    SafeConstructor,
):
    """YAML Loader with `!include` constructor and which can remember anchors
    between documents"""

    def __init__(self, stream):
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        RememberComposer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)
        SourceMappingConstructor.__init__(self)

    env_path_list: Optional[list] = None
    env_var_name = "TAVERN_INCLUDE"


def _get_include_dirs(loader):
    loader_list = [loader._root]

    if IncludeLoader.env_path_list is None:
        if IncludeLoader.env_var_name in os.environ:
            IncludeLoader.env_path_list = [
                os.path.expandvars(path_part)
                for path_part in os.environ[IncludeLoader.env_var_name].split(":")
            ]
        else:
            IncludeLoader.env_path_list = []

    return chain(loader_list, IncludeLoader.env_path_list)


def find_include(loader, node) -> str:
    """Locate an include file and return the abs path."""
    for directory in _get_include_dirs(loader):
        filename = os.path.abspath(
            os.path.join(directory, loader.construct_scalar(node))
        )
        if os.access(filename, os.R_OK):
            return filename

    raise BadSchemaError(
        "{} not found in include path: {}".format(
            loader.construct_scalar(node), [str(d) for d in _get_include_dirs(loader)]
        )
    )


def construct_include(loader, node):
    """Include file referenced at node."""

    filename = find_include(loader, node)
    extension = os.path.splitext(filename)[1].lstrip(".")

    if extension not in ("yaml", "yml", "json"):
        raise BadSchemaError(
            "Unknown filetype '{}' (included files must be in YAML format and end with .yaml or .yml)".format(
                filename
            )
        )

    return load_single_document_yaml(filename)


IncludeLoader.add_constructor("!include", construct_include)
IncludeLoader.add_constructor("!uuid", makeuuid)


class TypeSentinel(yaml.YAMLObject):
    """This is a sentinel for expecting a type in a response. Any value
    associated with these is going to be ignored - these are only used as a
    'hint' to the validator that it should expect a specific type in the
    response.
    """

    yaml_loader = IncludeLoader

    @staticmethod
    def constructor(_):
        raise NotImplementedError

    @classmethod
    def from_yaml(cls, loader, node) -> "TypeSentinel":
        return cls()

    def __str__(self) -> str:
        return f"<Tavern YAML sentinel for {self.constructor}>"

    @classmethod
    def to_yaml(cls, dumper, data) -> ScalarNode:
        node = yaml.nodes.ScalarNode(cls.yaml_tag, "", style=cls.yaml_flow_style)
        return node


class IntSentinel(TypeSentinel):
    yaml_tag = "!anyint"
    constructor = int


class FloatSentinel(TypeSentinel):
    yaml_tag = "!anyfloat"
    constructor = float


class StrSentinel(TypeSentinel):
    yaml_tag = "!anystr"
    constructor = str


class BoolSentinel(TypeSentinel):
    yaml_tag = "!anybool"
    constructor = bool


class ListSentinel(TypeSentinel):
    yaml_tag = "!anylist"
    constructor = list


class DictSentinel(TypeSentinel):
    yaml_tag = "!anydict"
    constructor = dict


@dataclasses.dataclass
class RegexSentinel(TypeSentinel):
    """Sentinel that matches a regex in a part of the response

    This shouldn't be used directly and instead one of the below match/fullmatch/search tokens will be used
    """

    constructor = str
    compiled: re.Pattern

    def __str__(self) -> str:
        return f"<Tavern Regex sentinel for {self.compiled.pattern}>"

    @property
    def yaml_tag(self):
        raise NotImplementedError

    @abstractmethod
    def passes(self, string):
        raise NotImplementedError

    @classmethod
    def from_yaml(cls, loader, node) -> "RegexSentinel":
        return cls(re.compile(node.value))


class _RegexMatchSentinel(RegexSentinel):
    yaml_tag = "!re_match"

    def passes(self, string) -> bool:
        return self.compiled.match(string) is not None


class _RegexFullMatchSentinel(RegexSentinel):
    yaml_tag = "!re_fullmatch"

    def passes(self, string) -> bool:
        return self.compiled.fullmatch(string) is not None


class _RegexSearchSentinel(RegexSentinel):
    yaml_tag = "!re_search"

    def passes(self, string) -> bool:
        return self.compiled.search(string) is not None


class AnythingSentinel(TypeSentinel):
    yaml_tag = "!anything"
    constructor = "anything"

    @classmethod
    def from_yaml(cls, loader, node):
        return ANYTHING

    def __deepcopy__(self, memo):
        """Return ANYTHING when doing a deep copy

        This is required because the checks in various parts of the code assume
        that ANYTHING is a singleton, but doing a deep copy creates a new object
        by default
        """
        return ANYTHING


# One instance of this (see above)
ANYTHING = AnythingSentinel()


class TypeConvertToken(yaml.YAMLObject):
    """This is a sentinel for something that should be converted to a different
    type. The rough load order is:

    1. Test data is loaded for schema validation
    2. Test data is dumped again so that pykwalify can read it (the actual
        values don't matter at all at this point, because we're just checking
        that the structure is correct)
    3. Test data is loaded and formatted

    So this preserves the actual value that the type should be up until the
    point that it actually needs to be formatted
    """

    yaml_loader = IncludeLoader

    @staticmethod
    def constructor(_):
        raise NotImplementedError

    def __init__(self, value) -> None:
        self.value = value

    @classmethod
    def from_yaml(cls, loader, node):
        value = loader.construct_scalar(node)

        try:
            # See if it's already a valid value (eg, if we do `!int "2"`)
            converted = cls.constructor(value)
        except ValueError:
            # If not (eg, `!int "{int_value:d}"`)
            return cls(value)
        else:
            return converted

    @classmethod
    def to_yaml(cls, dumper, data) -> ScalarNode:
        return yaml.nodes.ScalarNode(
            cls.yaml_tag, data.value, style=cls.yaml_flow_style
        )


class IntToken(TypeConvertToken):
    yaml_tag = "!int"
    constructor = int


class FloatToken(TypeConvertToken):
    yaml_tag = "!float"
    constructor = float


class StrToBoolConstructor:
    """Using `bool` as a constructor directly will evaluate all strings to `True`."""

    def __new__(cls, s: str) -> bool:  # type:ignore
        return strtobool(s)


class BoolToken(TypeConvertToken):
    yaml_tag = "!bool"
    constructor = StrToBoolConstructor


class StrToRawConstructor:
    """Used when we want to ignore brace formatting syntax"""

    def __new__(cls, s):
        return str(s.replace("{", "{{").replace("}", "}}"))


class RawStrToken(TypeConvertToken):
    yaml_tag = "!raw"
    constructor = StrToRawConstructor


class ForceIncludeToken(TypeConvertToken):
    """Magic tag that changes the way string formatting works"""

    yaml_tag = "!force_original_structure"

    @staticmethod
    def constructor(_):
        raise ValueError


class DeprecatedForceIncludeToken(ForceIncludeToken):
    """Old name for the above"""

    yaml_tag = "!force_format_include"

    @staticmethod
    def constructor(_):
        raise ValueError


# Sort-of hack to try and avoid future API changes
ApproxScalar = type(pytest.approx(1.0))


class ApproxSentinel(yaml.YAMLObject, ApproxScalar):  # type:ignore
    yaml_tag = "!approx"
    yaml_loader = IncludeLoader

    @classmethod
    def from_yaml(cls, loader, node) -> ApproxBase:
        try:
            val = float(node.value)
        except (ValueError, TypeError) as e:
            logger.error(
                "Could not coerce '%s' to a float for use with !approx",
                type(node.value),
            )
            raise BadSchemaError from e
        else:
            return pytest.approx(val)

    @classmethod
    def to_yaml(cls, dumper, data) -> ScalarNode:
        return yaml.nodes.ScalarNode(
            "!approx", str(data.expected), style=cls.yaml_flow_style
        )


# Apparently this isn't done automatically?
yaml.dumper.Dumper.add_representer(ApproxScalar, ApproxSentinel.to_yaml)


def load_single_document_yaml(filename: Union[str, os.PathLike]) -> dict:
    """
    Load a yaml file and expect only one document

    Args:
        filename: path to document

    Returns:
        content of file

    Raises:
        UnexpectedDocumentsError: If more than one document was in the file
    """

    with open(filename, encoding="utf-8") as fileobj:
        try:
            contents = yaml.load(fileobj, Loader=IncludeLoader)  # type:ignore # noqa
        except yaml.composer.ComposerError as e:
            msg = "Expected only one document in this file but found multiple"
            raise exceptions.UnexpectedDocumentsError(msg) from e

    return contents


def error_on_empty_scalar(self, mark):
    location = "{mark.name:s}:{mark.line:d} - column {mark.column:d}".format(mark=mark)
    error = "Error at {} - cannot define an empty value in test - either give it a value or explicitly set it to None".format(
        location
    )

    raise exceptions.BadSchemaError(error)



import contextlib
import json
import sys
import tempfile
from textwrap import dedent
from unittest.mock import Mock, patch

import _pytest
import pytest
import yaml
from box import Box

from tavern._core import exceptions
from tavern._core.dict_util import _check_and_format_values, format_keys
from tavern._core.loader import ForceIncludeToken
from tavern._core.pytest.item import YamlItem
from tavern._core.schema.extensions import validate_file_spec
from tavern._core.strict_util import (
    StrictLevel,
    StrictSetting,
    validate_and_parse_option,
)
from tavern.core import run
from tavern.helpers import validate_content, validate_pykwalify, validate_regex


class FakeResponse:
    def __init__(self, text):
        self.text = text
        self.headers = {"test_header": text}


class TestRegex:
    def test_regex_match(self):
        response = FakeResponse("abchelloabc")

        matched = validate_regex(response, "(?P<greeting>hello)")

        assert "greeting" in matched["regex"]

    def test_regex_no_match(self):
        response = FakeResponse("abchelloabc")

        with pytest.raises(exceptions.RegexAccessError):
            validate_regex(response, "(?P<greeting>hola)")

    def test_regex_match_header(self):
        response = FakeResponse("abchelloabc")

        matched = validate_regex(response, "(?P<greeting>hello)", header="test_header")

        assert "greeting" in matched["regex"]

    def test_regex_no_match_header(self):
        response = FakeResponse("abchelloabc")

        with pytest.raises(exceptions.RegexAccessError):
            validate_regex(response, "(?P<greeting>hola)", header="test_header")

    @pytest.mark.parametrize(
        "match",
        (
            (r"val(?P<num>\d)", "path1", "val1"),
            (r"val(?P<num>\d)", "path2[0]", "val2"),
            (r"val(?P<num>\d)", "path3.sub", "val3"),
        ),
    )
    def test_in_jmespath(self, match):
        response = FakeResponse(
            json.dumps({"path1": "val1", "path2": ["val2"], "path3": {"sub": "val3"}})
        )

        expression, path, expected = match

        result = validate_regex(response, expression, in_jmespath=path)

        assert result["regex"]["num"] == expected[-1]


class TestRunAlone:
    def test_run_calls_pytest(self):
        """This should just return from pytest.main()"""

        with patch("tavern.core.pytest.main") as pmock:
            run("abc")

        assert pmock.called

    def test_normal_args(self):
        with patch("tavern.core.pytest.main") as pmock:
            run(
                **{
                    "tavern_global_cfg": None,
                    "in_file": "kfdoskdof",
                    "tavern_http_backend": "requests",
                    "tavern_mqtt_backend": "paho-mqtt",
                    "tavern_strict": True,
                }
            )

        assert pmock.called

    def test_extra_args(self):
        with pytest.raises(TypeError):
            with patch("tavern.core.pytest.main") as pmock:
                run(
                    **{
                        "tavern_global_cfg": None,
                        "in_file": "kfdoskdof",
                        "tavern_http_backend": "requests",
                        "tavern_mqtt_backend": "paho-mqtt",
                        "tavern_strict": True,
                        "gfg": "2efsf",
                    }
                )

        assert not pmock.called


class TestOptionParsing:
    valid = [
        f"{section:s}:{setting:s}"
        for section in ("json", "headers", "redirect_params")
        for setting in ("on", "off")
    ]

    @pytest.mark.parametrize("optval", valid)
    def test_strictness_parsing_good(self, pytestconfig, optval):
        args = pytestconfig._parser.parse_known_args([f"--tavern-strict={optval}"])
        assert "tavern_strict" in args
        assert args.tavern_strict == [optval]


class TestTavernRepr:
    @pytest.fixture(name="fake_item")
    def fix_fake_item(self, request):
        item = YamlItem.from_parent(
            name="Fake Test Item", parent=request.node, spec={}, path="/tmp/hello"
        )
        return item

    @pytest.fixture(autouse=True, scope="session")
    def add_opts(self, pytestconfig):
        from tavern._core.pytest.hooks import pytest_addoption

        with contextlib.suppress(ValueError):
            pytest_addoption(pytestconfig._parser)

    def _make_fake_exc_info(self, exc_type):
        # Copied from pytest tests
        class FakeExcinfo(_pytest._code.ExceptionInfo):  # type:ignore
            pass

        try:
            raise exc_type
        except exc_type:
            excinfo = FakeExcinfo(sys.exc_info())

        return excinfo

    def test_not_called_for_normal_exception(self, fake_item):
        """Does not call tavern repr for non tavern errors"""
        fake_info = self._make_fake_exc_info(RuntimeError)

        with patch("tavern._core.pytest.item.ReprdError") as rmock:
            fake_item.repr_failure(fake_info)

        assert not rmock.called

    @pytest.mark.parametrize("ini_flag", [True, False])
    def test_not_called_for_badschema_tavern_exception_(self, fake_item, ini_flag):
        """Does not call taven repr for badschemerror - tavern repr gives no useful information in this case"""
        fake_info = self._make_fake_exc_info(exceptions.BadSchemaError)

        with patch.object(fake_item.config, "getini", return_value=ini_flag):
            with patch("tavern._core.pytest.item.ReprdError") as rmock:
                fake_item.repr_failure(fake_info)

        assert not rmock.called

    def test_not_called_ini(self, fake_item):
        """Enable ini flag, should call old style"""
        fake_info = self._make_fake_exc_info(exceptions.InvalidSettingsError)

        with patch.object(fake_item.config, "getini", return_value=True):
            with patch("tavern._core.pytest.item.ReprdError") as rmock:
                fake_item.repr_failure(fake_info)

        assert not rmock.called

    def test_not_called_cli(self, fake_item):
        """Enable cli flag, should call old style"""
        fake_info = self._make_fake_exc_info(exceptions.InvalidSettingsError)

        with patch.object(fake_item.config, "getoption", return_value=True):
            with patch("tavern._core.pytest.item.ReprdError") as rmock:
                fake_item.repr_failure(fake_info)

        assert not rmock.called


@pytest.fixture(name="nested_response")
def fix_nested_response():
    class response_content:
        content = {
            "top": {
                "Thing": "value",
                "float": 0.1,
                "nested": {"doubly": {"inner_value": "value", "inner_list": [1, 2, 3]}},
            },
            "an_integer": 123,
            "a_string": "abc",
            "a_bool": True,
        }

        def json(self):
            return self.content

    return response_content()


class TestContent:
    def test_correct_jmes_path(self, nested_response):
        comparisons = [
            {"jmespath": "top.Thing", "operator": "eq", "expected": "value"},
            {"jmespath": "an_integer", "operator": "eq", "expected": 123},
            {
                "jmespath": "top.nested.doubly.inner_list",
                "operator": "type",
                "expected": "list",
            },
        ]
        validate_content(nested_response, comparisons)
        assert True

    def test_incorrect_jmes_path(self, nested_response):
        comparisons = [{"jmespath": "userId", "operator": "eq", "expected": 1}]
        with pytest.raises(exceptions.JMESError):
            validate_content(nested_response, comparisons)

    def test_incorrect_value(self, nested_response):
        comparisons = [{"jmespath": "a_bool", "operator": "eq", "expected": False}]
        with pytest.raises(exceptions.JMESError):
            validate_content(nested_response, comparisons)


@pytest.mark.xfail
class TestPykwalifyExtension:
    def test_validate_schema_correct(self, nested_response):
        correct_schema = dedent(
            """
              type: map
              required: true
              mapping:
                top:
                  type: map
                  required: true
                  mapping:
                    Thing:
                      type: str
                    float:
                      type: float
                    nested:
                      type: any
                an_integer:
                  type: int
                a_string:
                  type: str
                a_bool:
                  type: bool
        """
        )

        validate_pykwalify(
            nested_response, yaml.load(correct_schema, Loader=yaml.SafeLoader)
        )

    def test_validate_schema_incorrect(self, nested_response):
        correct_schema = dedent(
            """
              type: seq
              required: true
              sequence:
                - type: str
        """
        )

        with pytest.raises(exceptions.BadSchemaError):
            validate_pykwalify(
                nested_response, yaml.load(correct_schema, Loader=yaml.SafeLoader)
            )


class TestCheckParseValues:
    @pytest.mark.parametrize("item", [yaml, yaml.load, yaml.SafeLoader])
    def test_warns_bad_type(self, item):
        with patch("tavern._core.dict_util.logger.warning") as wmock:
            _check_and_format_values("{fd}", Box({"fd": item}))

        wmock.assert_called_with(
            "Formatting '%s' will result in it being coerced to a string (it is a %s)",
            "fd",
            type(item),
        )

    @pytest.mark.parametrize(
        "item",
        [
            [134],
            {"a": 2},
        ],
    )
    def test_warns_bad_type_box(self, item):
        box = Box({"fd": item})
        with patch("tavern._core.dict_util.logger.warning") as wmock:
            _check_and_format_values("{fd}", box)

        wmock.assert_called_with(
            "Formatting '%s' will result in it being coerced to a string (it is a %s)",
            "fd",
            type(box["fd"]),
        )

    @pytest.mark.parametrize("item", [1, "a", 1.3, format_keys("{s}", {"s": 2})])
    def test_no_warn_good_type(self, item):
        with patch("tavern._core.dict_util.logger.warning") as wmock:
            _check_and_format_values("{fd}", Box({"fd": item}))

        assert not wmock.called

    def test_format_with_array_access(self):
        """Test accessing using array indexing format"""

        assert "hello" == _check_and_format_values(
            "{a.b.c[0].d}", Box({"a": {"b": {"c": [{"d": "hello"}]}}})
        )

    def test_format_with_dict_access(self):
        """Test accessing using dict indexing format"""

        assert "hello" == _check_and_format_values(
            "{a[b].c[0].d}", Box({"a": {"b": {"c": [{"d": "hello"}]}}})
        )


class TestFormatWithJson:
    @pytest.mark.parametrize(
        "item", [[134], {"a": 2}, yaml, yaml.load, yaml.SafeLoader]
    )
    def test_custom_format(self, item):
        """Can format everything"""
        val = format_keys(ForceIncludeToken("{fd}"), {"fd": item})

        assert val == item

    def test_bad_format_string_extra(self):
        """Extra things in format string"""
        with pytest.raises(exceptions.InvalidFormattedJsonError):
            format_keys(ForceIncludeToken("{fd}gg"), {"fd": "123"})

    def test_bad_format_string_conversion(self):
        """No format string"""
        with pytest.raises(exceptions.InvalidFormattedJsonError):
            format_keys(ForceIncludeToken(""), {"fd": "123"})

    def test_bad_format_string_multiple(self):
        """Multple format spec in string is disallowed"""
        with pytest.raises(exceptions.InvalidFormattedJsonError):
            format_keys(ForceIncludeToken("{a}{b}"), {"fd": "123"})


class TestCheckFileSpec:
    def _wrap_test_block(self, dowith):
        validate_file_spec({"files": dowith}, Mock(), Mock())

    def test_string_valid(self):
        with tempfile.NamedTemporaryFile() as tfile:
            self._wrap_test_block(tfile.name)

    def test_dict_valid(self):
        with tempfile.NamedTemporaryFile() as tfile:
            self._wrap_test_block({"file_path": tfile.name})

    def test_nonexistsnt_string(self):
        with pytest.raises(exceptions.BadSchemaError):
            self._wrap_test_block("kdsfofs")

    def nonexistent_dict(self):
        with pytest.raises(exceptions.BadSchemaError):
            self._wrap_test_block({"file_path": "gogfgl"})

    def extra_keys_dict(self):
        with pytest.raises(exceptions.BadSchemaError):
            self._wrap_test_block({"file_path": "gogfgl", "blop": 123})


class TestStrictUtils:
    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    @pytest.mark.parametrize("setting", ["on", "off"])
    def test_parse_option(self, section, setting):
        option = f"{section}:{setting}"
        match = validate_and_parse_option(option)

        assert match.section == section

        if setting == "on":
            assert match.is_on()
        else:
            assert not match.is_on()

    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    def test_unset_defaults(self, section):
        match = validate_and_parse_option(section)

        if section == "json":
            assert match.is_on()
        else:
            assert not match.is_on()

    @pytest.mark.parametrize("setting", ["true", "1", "hi", ""])
    def test_fails_bad_setting(self, setting):
        with pytest.raises(exceptions.InvalidConfigurationException):
            validate_and_parse_option(f"json:{setting}")

    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    def test_defaults_good(self, section):
        level = StrictLevel()

        if section == "json":
            assert level.option_for(section).is_on()
        else:
            assert not level.option_for(section).is_on()

    @pytest.mark.parametrize("section", ["true", "1", "hi", ""])
    def test_defaults_bad(self, section):
        level = StrictLevel()

        with pytest.raises(exceptions.InvalidConfigurationException):
            level.option_for(section)

    # These tests could be removed, they are testing implementation details...
    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    def test_set_on(self, section):
        level = StrictLevel.from_options([section + ":on"])

        assert level.option_for(section).setting == StrictSetting.ON
        assert level.option_for(section).is_on()

    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    def test_set_off(self, section):
        level = StrictLevel.from_options([section + ":off"])

        assert level.option_for(section).setting == StrictSetting.OFF
        assert not level.option_for(section).is_on()

    @pytest.mark.parametrize("section", ["json", "headers", "redirect_query_params"])
    def test_unset(self, section):
        level = StrictLevel.from_options([section])

        assert level.option_for(section).setting == StrictSetting.UNSET



from flask import Flask, jsonify, request

app = Flask(__name__)


@app.route("/double", methods=["POST"])
def double_number():
    r = request.get_json()

    try:
        number = r["number"]
    except (KeyError, TypeError):
        return jsonify({"error": "no number passed"}), 400

    try:
        double = int(number) * 2
    except ValueError:
        return jsonify({"error": "a number was not passed"}), 400

    return jsonify({"double": double}), 200



import dataclasses
import functools
import json
import logging
import warnings
from typing import Union

import grpc
from box import Box

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys, format_keys
from tavern._core.pytest.config import TestConfig
from tavern._plugins.grpc.client import GRPCClient
from tavern.request import BaseRequest

logger: logging.Logger = logging.getLogger(__name__)


def get_grpc_args(rspec: dict, test_block_config: TestConfig) -> dict:
    """Format GRPC request args"""

    fspec = format_keys(rspec, test_block_config.variables)

    # FIXME: Clarify 'json' and 'body' for grpc requests
    # FIXME 2: also it should allow proto text format. Maybe binary.
    if "json" in rspec:
        if "body" in rspec:
            raise exceptions.BadSchemaError(
                "Can only specify one of 'body' or 'json' in GRPC request"
            )

        fspec["body"] = json.dumps(fspec.pop("json"))

    return fspec


@dataclasses.dataclass
class WrappedFuture:
    response: Union[grpc.Call, grpc.Future]
    service_name: str


class GRPCRequest(BaseRequest):
    """Wrapper for a single GRPC request on a client

    Similar to RestRequest, publishes a single message.
    """

    _warned = False

    def __init__(
        self, client: GRPCClient, request_spec: dict, test_block_config: TestConfig
    ) -> None:
        if not self._warned:
            warnings.warn(
                "Tavern gRPC support is experimental and will be updated in a future release.",
                RuntimeWarning,
                stacklevel=0,
            )
            GRPCRequest._warned = True

        expected = {"host", "service", "body"}

        check_expected_keys(expected, request_spec)

        grpc_args = get_grpc_args(request_spec, test_block_config)

        self._prepared = functools.partial(client.call, **grpc_args)

        self._service_name = grpc_args.get("service", None)

        # Need to do this here because get_publish_args will modify the original
        # input, which we might want to use to format. No error handling because
        # all the error handling is done in the previous call
        self._original_request_vars = format_keys(
            request_spec, test_block_config.variables
        )

    def run(self) -> WrappedFuture:
        try:
            return WrappedFuture(
                response=self._prepared(), service_name=self._service_name
            )
        except ValueError as e:
            logger.exception("Error executing request")
            raise exceptions.GRPCRequestException from e

    @property
    def request_vars(self) -> Box:
        return Box(self._original_request_vars)



import contextlib
import logging
import logging.config
import os
import sqlite3

import paho.mqtt.client as paho
import yaml
from flask import Flask, g, jsonify, request

app = Flask(__name__)
application = app

DATABASE = os.environ.get("DB_NAME")


@contextlib.contextmanager
def get_client():
    mqtt_client = paho.Client(transport="websockets", client_id="server")
    mqtt_client.enable_logger()
    mqtt_client.username_pw_set(username="tavern", password="tavern")

    mqtt_client.connect(host="broker", port=9001)

    try:
        mqtt_client.loop_start()
        yield mqtt_client
    finally:
        mqtt_client.disconnect()
        mqtt_client.loop_stop()


def get_db():
    return sqlite3.connect(DATABASE)


def get_cached_db():
    db = getattr(g, "_database", None)
    if db is None:
        db = g._database = get_db()

    return db


def setup_logging():
    log_cfg = """
version: 1
disable_existing_loggers: true
formatters:
    fluent_fmt:
        (): fluent.handler.FluentRecordFormatter
        format:
            level: '%(levelname)s'
            where: '%(filename)s.%(lineno)d'
handlers:
    fluent:
        class: fluent.handler.FluentHandler
        formatter: fluent_fmt
        tag: server
        port: 24224
        host: fluent
loggers:
    paho:
        handlers:
            - fluent
        level: DEBUG
        propagate: true
    '':
        handlers:
            - fluent
        level: DEBUG
        propagate: true
"""

    as_dict = yaml.load(log_cfg, Loader=yaml.SafeLoader)
    logging.config.dictConfig(as_dict)

    logging.info("Logging set up")


@app.teardown_appcontext
def close_connection(exception):
    db = getattr(g, "_database", None)
    if db is not None:
        db.close()


@app.route("/send_mqtt_message", methods=["POST"])
def send_message():
    r = request.get_json()

    try:
        r["device_id"]
        r["payload"]
    except (KeyError, TypeError):
        return jsonify({"error": "missing key"}), 400

    topic = "/device/{}".format(r["device_id"])

    logging.debug("Publishing '%s' on '%s'", r["payload"], topic)

    try:
        with get_client() as mqtt_client:
            mqtt_client.publish(topic, r["payload"], r.get("qos", 0))
    except Exception:
        return jsonify({"error": topic}), 500

    return jsonify({"topic": topic}), 200


@app.route("/get_device_state", methods=["GET"])
def get_device():
    r = request.args

    try:
        r["device_id"]
    except (KeyError, TypeError):
        return jsonify({"error": "missing key"}), 400

    db = get_cached_db()

    with db:
        row = db.execute("SELECT * FROM devices_table WHERE device_id IS :device_id", r)

    try:
        status = next(row)[1]
    except StopIteration:
        return (
            jsonify(
                {"error": "could not find device with id {}".format(r["device_id"])}
            ),
            400,
        )

    onoff = "on" if status else "off"

    logging.info("Lights are %s", onoff)

    return jsonify({"lights": onoff})


@app.route("/create_device", methods=["PUT"])
def create_device():
    r = request.get_json(force=True)
    logging.error(r)

    try:
        r["device_id"]
    except (KeyError, TypeError):
        return jsonify({"error": "missing key device_id"}), 400

    db = get_cached_db()
    with db:
        row = db.execute(
            "SELECT device_id from devices_table where device_id is :device_id", r
        )

    try:
        r["clean"]
    except TypeError:
        return jsonify({"error": "checking for clean key"}), 500
    except KeyError:
        try:
            next(row)
        except StopIteration:
            pass
        else:
            return jsonify({"error": "device already exists"}), 400
    else:
        with db:
            db.execute("DELETE FROM devices_table")

    new_device = dict(lights_on=False, **r)

    logging.info("Creating new device: %s", new_device)

    with db:
        db.execute(
            "INSERT INTO devices_table (device_id, lights_on) VALUES (:device_id, :lights_on)",
            new_device,
        )

    return jsonify({"status": "created device {device_id}".format(**r)}), 201


@app.route("/reset", methods=["POST"])
def reset_db():
    db = get_cached_db()
    return _reset_db(db)


def _reset_db(db):
    with db:

        def attempt(query):
            with contextlib.suppress(Exception):
                db.execute(query)

        attempt("DELETE FROM devices_table")
        attempt(
            "CREATE TABLE devices_table (device_id TEXT NOT NULL, lights_on INTEGER NOT NULL)"
        )

    return "", 204


if __name__ == "__main__":
    setup_logging()
    db = get_db()
    _reset_db(db)



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from tavern._core.pytest.config import TestConfig


class TavernException(Exception):
    """Base exception

    Fields are internal and might change in future without warning

    Attributes:
        is_final: whether this exception came from a 'finally' block
        stage: stage that caused this issue
        test_block_config: config for stage
    """

    stage: Optional[dict]
    test_block_config: Optional["TestConfig"]
    is_final: bool = False


class BadSchemaError(TavernException):
    """Schema mismatch"""


class EvalError(TavernException):
    """Error parsing or running a simpleeval program"""


class TestFailError(TavernException):
    """Test failed somehow"""

    def __init__(self, msg, failures=None) -> None:
        super().__init__(msg)
        self.failures = failures or []


class KeyMismatchError(TavernException):
    """Mismatch found while validating keys in response"""


class UnexpectedKeysError(TavernException):
    """Unexpected keys used in request specification"""


class DuplicateKeysError(TavernException):
    """Duplicate key in request specification"""


class MissingKeysError(TavernException):
    """Missing key in request specification"""


class MissingFormatError(TavernException):
    """Tried to use a variable in a format string but it was not in the
    available variables
    """


class MissingSettingsError(TavernException):
    """Wanted to send an MQTT message but no settings were given"""


class MQTTError(TavernException):
    """Some kind of error returned from paho library"""


class MissingCookieError(TavernException):
    """Tried to use a cookie in a request that was not present in the session
    cookie jar
    """


class RestRequestException(TavernException):
    """Error making requests in RestRequest()"""


class GRPCRequestException(TavernException):
    """Error making requests in GRPCRequest()"""


class GRPCServiceException(TavernException):
    """Some kind of error when trying to get the gRPC service"""


class ProtoCompilerException(TavernException):
    """Some kind of error using protoc"""


class MQTTRequestException(TavernException):
    """Error making requests in MQTTRequest()"""


class MQTTTopicException(TavernException):
    """Internal (?) error with subscriptions"""


class MQTTTLSError(TavernException):
    """Error with TLS arguments to MQTT client"""


class PluginLoadError(TavernException):
    """Error loading a plugin"""


class InvalidExtFunctionError(TavernException):
    """Error loading an external function for validation/plugin use"""


class JMESError(TavernException):
    """Error in JMES matching"""


class InvalidStageReferenceError(TavernException):
    """Error loading stage reference"""


class DuplicateStageDefinitionError(TavernException):
    """Stage with the specified ID previously defined"""


class InvalidSettingsError(TavernException):
    """Configuration was passed incorrectly in some fashion"""


class KeySearchNotFoundError(TavernException):
    """Trying to search for a key in the response but was not found"""


class InvalidQueryResultTypeError(TavernException):
    """Searched for a value in data but it was not a 'simple' type"""


class UnexpectedDocumentsError(TavernException):
    """Multiple documents were found in a YAML file when only one was expected"""


class DuplicateCookieError(TavernException):
    """User tried to reuse a cookie from a previous request and override it in the same request"""


class InvalidConfigurationException(TavernException):
    """A configuration value (from the cli or the ini file) was invalid"""


class InvalidFormattedJsonError(TavernException):
    """Tried to use the magic json format tag in an invalid way"""


class MisplacedExtBlockException(TavernException):
    """Tried to use the '$ext' block in a place it is no longer valid to use it"""

    def __init__(self, block) -> None:
        super().__init__(
            "$ext function found in block {} - this has been moved to verify_response_with block - see documentation".format(
                block
            )
        )


class InvalidRetryException(TavernException):
    """Invalid spec for max_retries"""


class RegexAccessError(TavernException):
    """Error accessing a key via regex"""


class DuplicateStrictError(TavernException):
    """Tried to set stage strictness for multiple responses"""


class ConcurrentError(TavernException):
    """Error while processing concurrent future"""


class UnexpectedExceptionError(TavernException):
    """We expected a certain kind of exception in check_exception_raised but it was something
    else"""



import logging
import logging.config
import random

import pytest
import requests
import yaml

from tavern._core.pytest.item import YamlItem

logging_initialised = False


def setup_logging():
    log_cfg = """
version: 1
disable_existing_loggers: true
formatters:
    fluent_fmt:
        (): fluent.handler.FluentRecordFormatter
        format:
            level: '%(levelname)s'
            where: '%(filename)s.%(lineno)d'
    default:
        (): colorlog.ColoredFormatter
        format: "%(asctime)s [%(bold)s%(log_color)s%(levelname)s%(reset)s]: (%(bold)s%(name)s:%(lineno)d%(reset)s) %(message)s"
        style: "%"
        datefmt: "%X"
        log_colors:
            DEBUG:    cyan
            INFO:     green
            WARNING:  yellow
            ERROR:    red
            CRITICAL: red,bg_white
handlers:
    fluent:
        class: fluent.handler.FluentHandler
        formatter: fluent_fmt
        tag: tavern
        port: 24224
        host: localhost
        level: INFO
    stderr:
        class: colorlog.StreamHandler
        formatter: default
        level: DEBUG
loggers:
    paho: &log
        handlers:
            - stderr
        level: DEBUG
        propagate: False
    tavern:
        <<: *log

    tavern.mqtt: &reduced_log
        handlers:
            - stderr
            - fluent
        level: INFO
        propagate: False
    tavern.response.mqtt:
        <<: *reduced_log
    tavern.request.mqtt:
        <<: *reduced_log
"""

    as_dict = yaml.load(log_cfg, Loader=yaml.SafeLoader)
    logging.config.dictConfig(as_dict)

    logging.info("Logging set up")

    global logging_initialised
    logging_initialised = True


def pytest_runtest_setup(item):
    """Hack to get around pytest bug

    pytest doesn't appear to run 'autouse' fixtures unless the test being run is
    actually a normal python test, not for our custom tests - this runs
    setup_logging once for tavern tests
    """
    if isinstance(item, YamlItem) and not logging_initialised:
        setup_logging()

    return False


@pytest.fixture(scope="session", autouse=True)
def reset_devices():
    requests.post("http://localhost:5002/reset")


@pytest.fixture
def get_publish_topic(random_device_id):
    return "/device/{}/echo".format(random_device_id)


@pytest.fixture
def get_response_topic_suffix():
    return "response"


@pytest.fixture(scope="function", autouse=True)
def random_device_id():
    return str(random.randint(100, 10000))


@pytest.fixture(scope="function", autouse=True)
def random_device_id_2():
    return str(random.randint(100, 10000))



def strtobool(val: str) -> bool:
    """Copied and slightly modified from distutils as it's being removed in a future version of
    Python"""
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {val!r}")



import pytest

from tavern._core import exceptions
from tavern._core.schema.extensions import (
    validate_grpc_status_is_valid_or_list_of_names as validate_grpc,
)


class TestGrpcCodes:
    @pytest.mark.parametrize("code", ("UNAVAILABLE", "unavailable", "ok", 14, 0))
    def test_validate_grpc_valid_status(self, code):
        assert True is validate_grpc(code, None, None)
        assert True is validate_grpc([code], None, None)

    @pytest.mark.parametrize("code", (-1, "fo", "J", {"status": "OK"}))
    def test_validate_grpc_invalid_status(self, code):
        with pytest.raises(exceptions.BadSchemaError):
            assert False is validate_grpc(code, None, None)

        with pytest.raises(exceptions.BadSchemaError):
            assert False is validate_grpc([code], None, None)



import copy
import dataclasses
import logging
from importlib.util import find_spec
from typing import Any

from tavern._core.strict_util import StrictLevel

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class TavernInternalConfig:
    """Internal config that should be used only by tavern"""

    pytest_hook_caller: Any
    backends: dict


@dataclasses.dataclass(frozen=True)
class TestConfig:
    """Tavern configuration - there is a global config, then test-specific config, and
    finally stage-specific config, but they all use this structure

    Attributes:
        follow_redirects: whether the test should follow redirects
        variables: variables available for use in the stage
        strict: Strictness for test/stage
        stages: Any extra stages imported from other config files
    """

    variables: dict
    strict: StrictLevel
    follow_redirects: bool
    stages: list

    tavern_internal: TavernInternalConfig

    def copy(self) -> "TestConfig":
        """Returns a shallow copy of self"""
        return copy.copy(self)

    def with_new_variables(self) -> "TestConfig":
        """Returns a shallow copy of self but with the variables copied. This stops things being
        copied between tests. Can't use deepcopy because the variables might contain things that
        can't be pickled and hence can't be deep copied."""
        copied = self.copy()
        return dataclasses.replace(copied, variables=copy.copy(self.variables))

    def with_strictness(self, new_strict: StrictLevel) -> "TestConfig":
        """Create a copy of the config but with a new strictness setting"""
        return dataclasses.replace(self, strict=new_strict)

    @staticmethod
    def backends() -> list[str]:
        available_backends = ["http"]

        if has_module("paho.mqtt"):
            available_backends.append("mqtt")
        if has_module("grpc"):
            available_backends.append("grpc")

        logger.debug(f"available request backends: {available_backends}")

        return available_backends


def has_module(module: str) -> bool:
    try:
        return find_spec(module) is not None
    except ModuleNotFoundError:
        return False



import logging
from textwrap import dedent
from typing import Union

import yaml

try:
    from allure import attach, step
    from allure import attachment_type as at

    yaml_type = at.YAML
except ImportError:
    yaml_type = None

    def attach(*args, **kwargs) -> None:
        logger.debug("Not attaching anything as allure is not installed")

    def step(name):
        def call(step_func):
            return step_func

        return call


from tavern._core.formatted_str import FormattedString
from tavern._core.stage_lines import get_stage_lines, read_relevant_lines

logger: logging.Logger = logging.getLogger(__name__)


def prepare_yaml(val: Union[dict, set, list, tuple, str]) -> Union[dict, list, str]:
    """Sanitises the formatted string into a format safe for dumping"""
    if isinstance(val, dict):
        prepared = {}
        # formatted = {key: format_keys(val[key], box_vars) for key in val}
        for key in val:
            if isinstance(key, FormattedString):
                key = str(key)
            prepared[key] = prepare_yaml(val[key])

        return prepared
    elif isinstance(val, (list, tuple, set)):
        return [prepare_yaml(item) for item in val]
    elif isinstance(val, FormattedString):
        return str(val)

    return val


def attach_stage_content(stage: dict) -> None:
    first_line, last_line, _ = get_stage_lines(stage)

    code_lines = list(read_relevant_lines(stage, first_line, last_line))
    joined = dedent("\n".join(code_lines))
    attach_text(joined, "stage_yaml", yaml_type)


def attach_yaml(payload, name: str) -> None:
    prepared = prepare_yaml(payload)
    dumped = yaml.safe_dump(prepared)
    return attach_text(dumped, name, yaml_type)


def attach_text(payload, name: str, attachment_type=None) -> None:
    return attach(payload, name=name, attachment_type=attachment_type)


def wrap_step(allure_name: str, partial):
    return step(allure_name)(partial)



import pytest

from tavern._core.strict_util import StrictOption, StrictSetting, extract_strict_setting


@pytest.mark.parametrize(
    "strict", [True, StrictSetting.ON, StrictOption("json", StrictSetting.ON)]
)
def test_extract_strict_setting_true(strict):
    as_bool, as_setting = extract_strict_setting(strict)
    assert as_bool is True
    if isinstance(strict, StrictSetting):
        assert as_setting == strict
    if isinstance(strict, StrictOption):
        assert as_setting == strict.setting


@pytest.mark.parametrize(
    "strict",
    [
        False,
        StrictSetting.OFF,
        StrictSetting.LIST_ANY_ORDER,
        StrictSetting.UNSET,
        StrictOption("json", StrictSetting.OFF),
        None,
    ],
)
def test_extract_strict_setting_false(strict):
    as_bool, as_setting = extract_strict_setting(strict)
    assert as_bool is False
    if isinstance(strict, StrictSetting):
        assert as_setting == strict
    if isinstance(strict, StrictOption):
        assert as_setting == strict.setting



import copy
import dataclasses
import logging
import ssl
import threading
import time
from collections.abc import Mapping, MutableMapping
from queue import Empty, Full, Queue
from typing import Any, Optional, Union

import paho.mqtt.client as paho
from paho.mqtt.client import MQTTMessageInfo

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys

# MQTT error values
_err_vals = {
    -1: "MQTT_ERR_AGAIN",
    0: "MQTT_ERR_SUCCESS",
    1: "MQTT_ERR_NOMEM",
    2: "MQTT_ERR_PROTOCOL",
    3: "MQTT_ERR_INVAL",
    4: "MQTT_ERR_NO_CONN",
    5: "MQTT_ERR_CONN_REFUSED",
    6: "MQTT_ERR_NOT_FOUND",
    7: "MQTT_ERR_CONN_LOST",
    8: "MQTT_ERR_TLS",
    9: "MQTT_ERR_PAYLOAD_SIZE",
    10: "MQTT_ERR_NOT_SUPPORTED",
    11: "MQTT_ERR_AUTH",
    12: "MQTT_ERR_ACL_DENIED",
    13: "MQTT_ERR_UNKNOWN",
    14: "MQTT_ERR_ERRNO",
    15: "MQTT_ERR_QUEUE_SIZE",
}

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _Subscription:
    topic: str
    subscribed: bool = False

    # Arbitrary number, could just be 1 and only accept 1 message per stages
    # but we might want to raise an error if more than 1 message is received
    # during a test stage.
    queue: Queue = dataclasses.field(default_factory=lambda: Queue(maxsize=30))


def check_file_exists(key, filename) -> None:
    try:
        with open(filename, encoding="utf-8"):
            pass
    except OSError as e:
        raise exceptions.MQTTTLSError(f"Couldn't load '{key}' from '{filename}'") from e


def _handle_tls_args(
    tls_args: MutableMapping,
) -> Optional[Mapping]:
    """Make sure TLS options are valid"""

    if not tls_args:
        return None

    if "enable" in tls_args and not tls_args["enable"]:
        # if enable=false, return immediately
        return None

    _check_and_update_common_tls_args(tls_args, ["certfile", "keyfile"])

    return tls_args


def _handle_ssl_context_args(
    ssl_context_args: MutableMapping,
) -> Optional[Mapping]:
    """Make sure SSL Context options are valid"""
    if not ssl_context_args:
        return None

    _check_and_update_common_tls_args(
        ssl_context_args, ["certfile", "keyfile", "cafile"]
    )

    return ssl_context_args


def _check_and_update_common_tls_args(
    tls_args: MutableMapping, check_file_keys: list[str]
) -> None:
    """Checks common args between ssl/tls args"""

    # could be moved to schema validation stage
    for key in check_file_keys:
        if key in tls_args:
            check_file_exists(key, tls_args[key])

    if "keyfile" in tls_args and "certfile" not in tls_args:
        raise exceptions.MQTTTLSError(
            "If specifying a TLS keyfile, a certfile also needs to be specified"
        )

    if "cert_reqs" in tls_args:
        tls_args["cert_reqs"] = getattr(ssl, tls_args["cert_reqs"])

    try:
        tls_args["tls_version"] = getattr(ssl, tls_args["tls_version"])
    except AttributeError as e:
        raise exceptions.MQTTTLSError(
            "Error getting TLS version from "
            "ssl module - ssl module had no attribute '{}'. Check the "
            "documentation for the version of python you're using to see "
            "if this a valid option.".format(tls_args["tls_version"])
        ) from e
    except KeyError:
        pass


class MQTTClient:
    def __init__(self, **kwargs) -> None:
        expected_blocks = {
            "client": {
                "client_id",
                "clean_session",
                # Can't really use this easily...
                # "userdata",
                # Force mqttv311 - fix if this becomes an issue
                # "protocol",
                "transport",
            },
            "connect": {"host", "port", "keepalive", "timeout"},
            "tls": {
                "enable",
                "ca_certs",
                "cert_reqs",
                "certfile",
                "keyfile",
                "tls_version",
                "ciphers",
            },
            "auth": {"username", "password"},
            "ssl_context": {
                "ca_certs",
                "certfile",
                "keyfile",
                "password",
                "tls_version",
                "ciphers",
                "alpn_protocols",
            },
        }

        sanitised_kwargs = copy.deepcopy(kwargs)
        if auth := kwargs.get("auth"):
            if "password" in auth:
                sanitised_kwargs["auth"]["password"] = "******"  # noqa

        logger.debug("Initialising MQTT client with %s", sanitised_kwargs)

        # check main block first
        check_expected_keys(expected_blocks.keys(), kwargs)

        # then check constructor/connect/tls_set args
        self._client_args = kwargs.pop("client", {})
        check_expected_keys(expected_blocks["client"], self._client_args)

        self._connect_args = kwargs.pop("connect", {})
        check_expected_keys(expected_blocks["connect"], self._connect_args)

        self._auth_args = kwargs.pop("auth", {})
        check_expected_keys(expected_blocks["auth"], self._auth_args)

        if "host" not in self._connect_args:
            msg = "Need 'host' in 'connect' block for mqtt"
            raise exceptions.MissingKeysError(msg)

        self._connect_timeout = self._connect_args.pop("timeout", 3)

        # If there is any tls or ssl_context kwarg, configure tls encryption
        file_tls_args = kwargs.pop("tls", {})
        file_ssl_context_args = kwargs.pop("ssl_context", {})

        if file_tls_args and file_ssl_context_args:
            msg = (
                "'tls' and 'ssl_context' are both specified but are mutually exclusive"
            )
            raise exceptions.MQTTTLSError(msg)

        check_expected_keys(expected_blocks["tls"], file_tls_args)
        self._tls_args = _handle_tls_args(file_tls_args)
        logger.debug("TLS is %s", "enabled" if self._tls_args else "disabled")

        # If there is any SSL kwarg, enable tls through the SSL context
        check_expected_keys(expected_blocks["ssl_context"], file_ssl_context_args)
        self._ssl_context_args = _handle_ssl_context_args(file_ssl_context_args)

        logger.debug("Paho client args: %s", self._client_args)
        self._client = paho.Client(**self._client_args)
        self._client.enable_logger()

        if self._auth_args:
            logger.debug("authenticating as '%s'", self._auth_args.get("username"))
            self._client.username_pw_set(**self._auth_args)

        self._client.on_message = self._on_message
        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect
        self._client.on_connect_fail = self._on_connect_fail
        self._client.on_socket_open = self._on_socket_open
        self._client.on_socket_close = self._on_socket_close

        if self._tls_args:
            try:
                self._client.tls_set(**self._tls_args)
            except ValueError as e:
                # tls_set only raises ValueErrors directly
                raise exceptions.MQTTTLSError("Unexpected error enabling TLS") from e
            except ssl.SSLError as e:
                # incorrect cipher, etc.
                raise exceptions.MQTTTLSError(
                    "Unexpected SSL error enabling TLS"
                ) from e

        if self._ssl_context_args:
            # Create SSLContext object
            tls_version = self._ssl_context_args.get("tls_version")
            if tls_version is None:
                # If the python version supports it, use highest TLS version automatically
                if hasattr(ssl, "PROTOCOL_TLS_CLIENT"):
                    tls_version = ssl.PROTOCOL_TLS_CLIENT
                elif hasattr(ssl, "PROTOCOL_TLS"):
                    tls_version = ssl.PROTOCOL_TLS
                else:
                    tls_version = ssl.PROTOCOL_TLSv1_2
            ca_certs = self._ssl_context_args.get("cert_reqs")
            context = ssl.create_default_context(cafile=ca_certs)

            certfile = self._ssl_context_args.get("certfile")
            keyfile = self._ssl_context_args.get("keyfile")
            password = self._ssl_context_args.get("password")

            # Configure context
            if certfile is not None:
                context.load_cert_chain(certfile, keyfile, password)

            cert_reqs = self._ssl_context_args.get("cert_reqs")
            if cert_reqs == ssl.CERT_NONE and hasattr(context, "check_hostname"):
                context.check_hostname = False

            context.verify_mode = ssl.CERT_REQUIRED if cert_reqs is None else cert_reqs

            if ca_certs is not None:
                context.load_verify_locations(ca_certs)
            else:
                context.load_default_certs()

            ciphers = self._ssl_context_args.get("ciphers")
            if ciphers is not None:
                context.set_ciphers(ciphers)

            alpn_protocols = self._ssl_context_args.get("alpn_protocols")
            if alpn_protocols is not None:
                context.set_alpn_protocols(alpn_protocols)

            self._client.tls_set_context(context)

            if cert_reqs != ssl.CERT_NONE:
                # Default to secure, sets context.check_hostname attribute
                # if available
                self._client.tls_insecure_set(False)
            else:
                # But with ssl.CERT_NONE, we can not check_hostname
                self._client.tls_insecure_set(True)

        # Topics to subscribe to - mapping of subscription message id to subscription object
        self._subscribed: dict[int, _Subscription] = {}
        # Lock to ensure there is no race condition when subscribing
        self._subscribe_lock = threading.RLock()
        # callback
        self._client.on_subscribe = self._on_subscribe

        # Mapping of topic -> subscription id, for indexing into self._subscribed
        self._subscription_mappings: dict[str, int] = {}
        self._userdata = {
            "_subscription_mappings": self._subscription_mappings,
            "_subscribed": self._subscribed,
        }
        self._client.user_data_set(self._userdata)

    @staticmethod
    def _on_message(
        client, userdata: Mapping[str, Any], message: paho.MQTTMessage
    ) -> None:
        """Add any messages received to the queue

        Todo:
            If the queue is faull trigger an error in main thread somehow
        """

        logger.info("Received mqtt message on %s", message.topic)

        try:
            for sub_topic, sub_id in userdata["_subscription_mappings"].items():
                if paho.topic_matches_sub(sub_topic, message.topic):
                    userdata["_subscribed"][sub_id].queue.put(message)
                    break
            else:
                raise exceptions.MQTTTopicException(
                    f"Message received on unregistered topic: {message.topic}"
                )
        except Full:
            logger.exception("message queue full")

    @staticmethod
    def _on_connect(client, userdata, flags, rc: int) -> None:
        logger.debug(
            "Client '%s' connected to the broker with result code '%s'",
            client._client_id.decode(),
            paho.connack_string(rc),
        )

    @staticmethod
    def _on_disconnect(client, userdata, rc: int) -> None:
        if rc == paho.CONNACK_ACCEPTED:
            logger.debug(
                "Client '%s' successfully disconnected from the broker with result code '%s'",
                client._client_id.decode(),
                paho.connack_string(rc),
            )
        else:
            logger.warning(
                "Client %s failed to disconnect cleanly due to %s, possibly from a network error",
                client._client_id.decode(),
                paho.connack_string(rc),
            )

    @staticmethod
    def _on_connect_fail(client, userdata) -> None:
        logger.error(
            "Failed to connect client '%s' to the broker", client._client_id.decode()
        )

    @staticmethod
    def _on_socket_open(client, userdata, socket) -> None:
        logger.debug("MQTT socket opened")

    @staticmethod
    def _on_socket_close(client, userdata, socket) -> None:
        logger.debug("MQTT socket closed")

    def message_received(
        self, topic: str, timeout: Union[float, int] = 1
    ) -> Optional[paho.MQTTMessage]:
        """Check that a message is in the message queue

        Args:
            topic: topic to fetch message for
            timeout: How long to wait before signalling that the message
                was not received.

        Returns:
            the message, if one was received, otherwise None

        Todo:
            Allow regexes for topic names? Better validation for mqtt payloads
        """

        try:
            with self._subscribe_lock:
                queue = self._subscribed[self._subscription_mappings[topic]].queue
        except KeyError as e:
            raise exceptions.MQTTTopicException(f"Unregistered topic: {topic}") from e

        try:
            msg = queue.get(block=True, timeout=timeout)
        except Empty:
            logger.error("Message not received after %d seconds", timeout)
            return None

        return msg

    def publish(
        self,
        topic: str,
        payload: Optional[Any] = None,
        qos: Optional[int] = None,
        retain: Optional[bool] = False,
    ) -> MQTTMessageInfo:
        """publish message using paho library"""
        self._wait_for_subscriptions()

        logger.debug("Publishing on '%s'", topic)

        kwargs = {}
        if qos is not None:
            kwargs["qos"] = qos
        if retain is not None:
            kwargs["retain"] = retain
        msg = self._client.publish(topic, payload, **kwargs)

        # Wait for 2*connect timeout which should be plenty to publish the message even with qos 2
        # TODO: configurable
        try:
            msg.wait_for_publish(self._connect_timeout * 2)
        except (RuntimeError, ValueError) as e:
            raise exceptions.MQTTError("could not publish message") from e

        if not msg.is_published():
            raise exceptions.MQTTError(
                "err {:s}: {:s}".format(
                    _err_vals.get(msg.rc, "unknown"), paho.error_string(msg.rc)
                )
            )

        return msg

    def _wait_for_subscriptions(self) -> None:
        """Wait for all pending subscriptions to finish"""
        logger.debug("Checking subscriptions")

        def not_finished_subscribing_to():
            """Get topic names for topics which have not finished subcribing to"""
            return [i.topic for i in self._subscribed.values() if not i.subscribed]

        to_wait_for = not_finished_subscribing_to()

        if to_wait_for:
            elapsed = 0.0
            while elapsed < self._connect_timeout:
                # TODO
                # configurable?
                time.sleep(0.25)
                elapsed += 0.25

                to_wait_for = not_finished_subscribing_to()

                if not to_wait_for:
                    break

                logger.debug(
                    "Not finished subscribing to '%s' after %.2f seconds",
                    to_wait_for,
                    elapsed,
                )

            if to_wait_for:
                logger.warning(
                    "Did not finish subscribing to '%s' before publishing - going ahead anyway",
                    to_wait_for,
                )

        if not to_wait_for:
            logger.debug("Finished subscribing to all topics")

    def subscribe(self, topic: str, *args, **kwargs) -> None:
        """Subscribe to topic

        should be called for every expected message in mqtt_response
        """
        logger.debug("Subscribing to topic '%s'", topic)

        (status, mid) = self._client.subscribe(topic, *args, **kwargs)

        if status == 0:
            with self._subscribe_lock:
                self._subscription_mappings[topic] = mid
                self._subscribed[mid] = _Subscription(topic)
        else:
            raise exceptions.MQTTError(
                f"Error subscribing to '{topic}' (err code {status})"
            )

    def unsubscribe_all(self) -> None:
        """Unsubscribe from all topics"""
        with self._subscribe_lock:
            for subscription in self._subscribed.values():
                self._client.unsubscribe(subscription.topic)

    def _on_subscribe(self, client, userdata, mid: int, granted_qos) -> None:
        with self._subscribe_lock:
            if mid in self._subscribed:
                self._subscribed[mid].subscribed = True
                logger.debug(
                    "Successfully subscribed to '%s'", self._subscribed[mid].topic
                )
            else:
                logger.debug("Only tracking: %s", self._subscribed.keys())
                logger.warning(
                    "Got SUBACK message with mid '%s', but did not recognise that mid - will try later",
                    mid,
                )

    def __enter__(self) -> "MQTTClient":
        logger.debug("Connecting to %s", self._connect_args)

        self._client.connect_async(**self._connect_args)
        self._client.loop_start()

        elapsed = 0.0

        while elapsed < self._connect_timeout:
            if self._client.is_connected():
                logger.debug("Connected to broker at %s", self._connect_args["host"])
                return self
            else:
                logger.debug("Not connected after %s seconds - waiting", elapsed)

            # TODO
            # configurable?
            time.sleep(0.25)
            elapsed += 0.25

        self._disconnect()
        logger.error(
            "Could not connect to broker after %s seconds", self._connect_timeout
        )
        raise exceptions.MQTTError

    def __exit__(self, *args) -> None:
        self._disconnect()

    def _disconnect(self) -> None:
        self._client.disconnect()
        self._client.loop_stop()



import functools
import importlib
import logging
from collections.abc import Iterable, Mapping
from typing import Any, Callable, Optional

from tavern._core import exceptions

from .dict_util import deep_dict_merge


def is_ext_function(block: Any) -> bool:
    """
    Whether the given object is an ext function block

    Args:
        block: Any object

    Returns:
        If it is an ext function style dict
    """
    return isinstance(block, dict) and block.get("$ext", None) is not None


def get_pykwalify_logger(module: Optional[str]) -> logging.Logger:
    """Get logger for this module

    Have to do it like this because the way that pykwalify load extension
    modules means that getting the logger the normal way just result sin it
    trying to get the root logger which won't log correctly

    Args:
        module: name of module to get logger for

    Returns:
        logger for given module
    """
    return logging.getLogger(module)


def _getlogger() -> logging.Logger:
    """Get logger for this module"""
    return get_pykwalify_logger("tavern._core.extfunctions")


def import_ext_function(entrypoint: str) -> Callable:
    """Given a function name in the form of a setuptools entry point, try to
    dynamically load and return it

    Args:
        entrypoint: setuptools-style entrypoint in the form
            module.submodule:function

    Returns:
        function loaded from entrypoint

    Raises:
        InvalidExtFunctionError: If the module or function did not exist
    """
    logger = _getlogger()

    try:
        module, funcname = entrypoint.split(":")
    except ValueError as e:
        msg = "Expected entrypoint in the form module.submodule:function"
        logger.exception(msg)
        raise exceptions.InvalidExtFunctionError(msg) from e

    try:
        imported = importlib.import_module(module)
    except ImportError as e:
        msg = f"Error importing module {module}"
        logger.exception(msg)
        raise exceptions.InvalidExtFunctionError(msg) from e

    try:
        function = getattr(imported, funcname)
    except AttributeError as e:
        msg = f"No function named {funcname} in {module}"
        logger.exception(msg)
        raise exceptions.InvalidExtFunctionError(msg) from e

    return function


def get_wrapped_response_function(ext: Mapping) -> Callable:
    """Wraps a ext function with arguments given in the test file

    This is similar to functools.wrap, but this makes sure that 'response' is
    always the first argument passed to the function

    Args:
        ext: $ext function dict with function, extra_args, and
            extra_kwargs to pass

    Returns:
        Wrapped function
    """

    func, args, kwargs = _get_ext_values(ext)

    @functools.wraps(func)
    def inner(response):
        result = func(response, *args, **kwargs)
        _getlogger().debug("Result of calling '%s': '%s'", func, result)
        return result

    inner.func = func  # type: ignore

    return inner


def get_wrapped_create_function(ext: Mapping) -> Callable:
    """Same as get_wrapped_response_function, but don't require a response"""

    func, args, kwargs = _get_ext_values(ext)

    @functools.wraps(func)
    def inner():
        result = func(*args, **kwargs)
        _getlogger().debug("Result of calling '%s': '%s'", func, result)
        return result

    inner.func = func  # type: ignore

    return inner


def _get_ext_values(ext: Mapping) -> tuple[Callable, Iterable, Mapping]:
    if not isinstance(ext, Mapping):
        raise exceptions.InvalidExtFunctionError(
            f"ext block should be a dict, but it was a {type(ext)}"
        )

    args = ext.get("extra_args") or ()
    kwargs = ext.get("extra_kwargs") or {}
    try:
        func = import_ext_function(ext["function"])
    except KeyError as e:
        raise exceptions.BadSchemaError(
            "No function specified in external function block"
        ) from e

    return func, args, kwargs


def update_from_ext(request_args: dict, keys_to_check: list[str]) -> None:
    """
    Updates the request_args dict with any values from external functions

    Args:
        request_args: dictionary of request args
        keys_to_check: list of keys in request to possibly update from
    """

    new_args = {}
    logger = _getlogger()

    for key in keys_to_check:
        try:
            block = request_args[key]
        except KeyError:
            logger.debug("No %s block", key)
            continue

        try:
            pop = block.pop("$ext")
        except (KeyError, AttributeError, TypeError):
            logger.debug("No ext functions in %s block", key)
            continue

        func = get_wrapped_create_function(pop)
        new_args[key] = func()

    merged_args = deep_dict_merge(request_args, new_args)

    request_args.update(**merged_args)



import datetime
import functools

import jwt
from flask import Flask, jsonify, request

app = Flask(__name__)

SECRET = "CGQgaG7GYvTcpaQZqosLy5"
DATABASE = "/tmp/test_db"
SERVERNAME = "testserver"


@app.route("/login", methods=["POST"])
def login():
    r = request.get_json()

    if r["user"] != "test-user" or r["password"] != "correct-password":
        return jsonify({"error": "Incorrect username/password"}), 401

    payload = {
        "sub": "test-user",
        "aud": SERVERNAME,
        "exp": datetime.datetime.utcnow() + datetime.timedelta(hours=1),
    }

    token = jwt.encode(payload, SECRET, algorithm="HS256")

    return jsonify({"token": token})


def requires_jwt(endpoint):
    """Makes sure a jwt is in the request before accepting it"""

    @functools.wraps(endpoint)
    def check_auth_call(*args, **kwargs):
        token = request.headers.get("Authorization")

        # check token is present
        if not token:
            return jsonify({"error": "No token"}), 401

        token_type, token = token.split(" ")

        if token_type.lower() != "bearer":
            return jsonify({"error": "Wrong token type"}), 401

        try:
            jwt.decode(token, SECRET, audience=SERVERNAME, algorithms=["HS256"])
        except:
            return jsonify({"error": "Invalid token"}), 401

        return endpoint(*args, **kwargs)

    return check_auth_call


@app.route("/ping", methods=["GET"])
@requires_jwt
def ping():
    return jsonify({"data": "pong"}), 200


@app.route("/hello/<name>", methods=["GET"])
@requires_jwt
def hello(name):
    return jsonify({"data": "Hello, {}".format(name)}), 200



from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class DummyRequest(_message.Message):
    __slots__ = ["request_id"]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: int
    def __init__(self, request_id: _Optional[int] = ...) -> None: ...

class DummyResponse(_message.Message):
    __slots__ = ["response_id"]
    RESPONSE_ID_FIELD_NUMBER: _ClassVar[int]
    response_id: int
    def __init__(self, response_id: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: test_services.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13test_services.proto\x12\x0ftavern.tests.v1\x1a\x1bgoogle/protobuf/empty.proto\"\"\n\x0c\x44ummyRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\x05\"$\n\rDummyResponse\x12\x13\n\x0bresponse_id\x18\x01 \x01(\x05\x32\x94\x01\n\x0c\x44ummyService\x12\x37\n\x05\x45mpty\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\x12K\n\nSimpleTest\x12\x1d.tavern.tests.v1.DummyRequest\x1a\x1e.tavern.tests.v1.DummyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'test_services_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DUMMYREQUEST']._serialized_start=69
  _globals['_DUMMYREQUEST']._serialized_end=103
  _globals['_DUMMYRESPONSE']._serialized_start=105
  _globals['_DUMMYRESPONSE']._serialized_end=141
  _globals['_DUMMYSERVICE']._serialized_start=144
  _globals['_DUMMYSERVICE']._serialized_end=292
# @@protoc_insertion_point(module_scope)
